#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from functools import partial
import numpy as np
from pricelib.common.utilities.enums import StatusType
from pricelib.common.date_handler import CN_CALENDAR, AnnualDays
from .autocallable_base import AutocallableBase


class ButterflySnowball(AutocallableBase):
    """蝶变雪球结构产品类"""

    def __init__(self, s0, barrier_out, barrier_in, coupon_out1, coupon_out2, coupon_out3, coupon_div=None, lock_term=3,
                 maturity=None, start_date=None, end_date=None, obs_dates=None, pay_dates=None, parti_in=1,
                 margin_lvl=1, coupon_stair_ends=(12, 24, 36), status=StatusType.NoTouch, engine=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 stoch_process=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """继承自动赎回基类AutocallableBase的参数，详见AutocallableBase的__init__方法
        Args: 以下为区别于经典雪球的参数
            coupon_out1: float，第一阶段敲出票息，百分比，年化
            coupon_out2: float，第二阶段敲出票息，百分比，年化
            coupon_out3: float，第三阶段敲出票息，百分比，年化
            coupon_stair_ends: Tuple(int), 分段票息节点, 通常为(12, 24, 36), 代表每个阶段到第几个月末为止
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, status=status,
                         annual_days=annual_days, parti_in=parti_in, margin_lvl=margin_lvl,
                         t_step_per_year=t_step_per_year, engine=engine, stoch_process=stoch_process,
                         s=s, r=r, q=q, vol=vol, engine_method=engine_method)
        len_obs_dates = len(self.obs_dates.date_schedule)

        self.barrier_out = np.ones(len_obs_dates) * barrier_out
        self.barrier_in = np.ones(len_obs_dates) * barrier_in
        self.coupon_stair_ends = coupon_stair_ends
        coupon_out = np.ones(len_obs_dates + lock_term - 1)
        coupon_out[:coupon_stair_ends[0]] = coupon_out1
        coupon_out[coupon_stair_ends[0]:coupon_stair_ends[1]] = coupon_out2
        coupon_out[coupon_stair_ends[1]:coupon_stair_ends[2]] = coupon_out3
        self.coupon_out = coupon_out[lock_term - 1:]
        self.coupon_div = coupon_div if coupon_div is not None else coupon_out[-1]
        self.parti_out = 0
        self.strike_upper = s0
        self.strike_lower = 0
        # 反算指定阶段的敲出票息(默认为nth=1)
        self.calc_coupon = partial(self.calc_param, reset_param_method=self.reset_coupon_nth)

    def __repr__(self):
        """返回期权的描述"""
        return "蝶式雪球"

    def reset_coupon_nth(self, coupon: float, nth: int = 1, coupon_div=None, *args, **kwargs):
        """重新设置指定分段的敲出票息
        Args:
            coupon: float, 重设票息值
            coupon_div: float, 是否指定了红利票息的值
            nth: int, 修改的是第几个阶段的票息
        """
        if nth == 1:
            self.coupon_out[:self.coupon_stair_ends[0] - self.lock_term + 1] = coupon
        else:
            self.coupon_out[self.coupon_stair_ends[nth - 2] - self.lock_term + 1:
                            self.coupon_stair_ends[nth - 1] - self.lock_term + 1] = coupon
        # 如果反算的是最后一段票息，并且没有指定红利票息，同时更新红利票息的值
        if (len(self.coupon_stair_ends) == nth) and (coupon_div is None):
            self.coupon_div = coupon



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .term_structures import *
from .volmodels import *
from .processes import *
from .date_handler import *
from .utilities.enums import *
from .utilities import logging, set_logging_handlers, SimpleQuote
from .pricing_engine_base import McReusablePaths



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from examples.digital import digital_option_demo, double_touch_notouch_demo


def test_digital_demo():
    expected_discrete = [
        ['离散观察到期支付欧式Call二元期权', 0.7813384345830714, 0.7800421042175157, 0.7722835839041402, 0.812813152699443,
         0.7891833475907255],
        ['离散观察到期支付欧式Put二元期权', 1.2726911914338912, 1.2630840104230847, 1.2731902121549123, 1.2895615455303617,
         1.2913388609349215],
        ['离散观察到期支付美式Call二元期权', 1.6766870002507048, 1.6744733936099303, 1.657521761793372, 1.6688088109426866,
         1.670733761754569],
        ['离散观察到期支付美式Put二元期权', 2.1376442967454254, 2.134676670727452, 2.1407896427877984, 2.0890054574578123,
         2.1428078891800553],
        ['离散观察立即支付美式Call二元期权', 1.6909576756985945, 1.688528407429802, 1.6715212118040825, 1.6828563953693552,
         1.6849260212121293],
        ['离散观察立即支付美式Put二元期权', 2.1530488132491263, 2.1497975527515845, 2.1561525555229144, 2.1037244495819074,
         2.1582634329368275]]
    expected_continuous = [['连续观察到期支付欧式Call二元期权', 0.7813384345830714, 0.789183],
                           ['连续观察到期支付欧式Put二元期权', 1.2726911914338912, 1.291339],
                           ['连续观察到期支付美式Call二元期权', 1.8021622711412713, 1.799008],
                           ['连续观察到期支付美式Put二元期权', 2.269883592086009, 2.263541],
                           ['连续观察立即支付美式Call二元期权', 1.817889647827623, 1.814803],
                           ['连续观察立即支付美式Put二元期权', 2.286636690933101, 2.280384]]
    expected_discrete_df = pd.DataFrame(expected_discrete, columns=['期权类型    ', '闭式解(每日观察)', 'MonteCarlo',
                                                                    'Quadrature', '二叉树', 'PDE'])
    expected_continuous_df = pd.DataFrame(expected_continuous, columns=['期权类型    ', '闭式解(连续观察)', 'PDE'])
    df1, df2 = digital_option_demo.run()
    assert_frame_equal(df1, expected_discrete_df, check_dtype=True)
    assert_frame_equal(df2, expected_continuous_df, check_dtype=True)


def test_double_touch_notouch_demo():
    expected_discrete = [['到期支付二元凹式(欧式双接触期权)', np.nan, 3.084881, 3.123992],
                         ['到期支付二元凸式(欧式双不接触期权)', np.nan, 6.717105, 6.678003],
                         ['到期支付离散观察美式双触碰二元期权', 5.766874556177397, 5.759843, 5.748472],
                         ['到期支付离散观察美式双不触碰二元期权', 4.035112176890156, 4.042143, 4.05352],
                         ['立即支付离散观察美式双触碰二元期权', np.nan, 5.813134, 5.801601],
                         ['立即支付离散观察美式双不触碰二元期权', np.nan, 4.042143, 4.05352]]

    expected_continuous = [['到期支付二元凹式(欧式双接触期权)', np.nan, 3.123992],
                           ['到期支付二元凸式(欧式双不接触期权)', np.nan, 6.678003],
                           ['到期支付连续观察美式双触碰二元期权', 6.095226932808097, 6.083505],
                           ['到期支付连续观察美式双不触碰二元期权', 3.706759800259456, 3.718488],
                           ['立即支付连续观察美式双触碰二元期权', np.nan, 6.141737],
                           ['立即支付连续观察美式双不触碰二元期权', np.nan, 3.718488]]

    expected_discrete_df = pd.DataFrame(expected_discrete,
                                        columns=['期权类型    ', '闭式解(每日观察)', 'MonteCarlo', 'PDE'])
    expected_continuous_df = pd.DataFrame(expected_continuous, columns=['期权类型    ', '闭式解(连续观察)', 'PDE'])
    df1, df2 = double_touch_notouch_demo.run()
    assert_frame_equal(df1, expected_discrete_df, check_dtype=True)
    assert_frame_equal(df2, expected_continuous_df, check_dtype=True)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import bisect
import numpy as np
import datetime
from pricelib.common.utilities.enums import RandsMethod, LdMethod, StatusType, ExerciseType
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.date_handler import global_evaluation_date


class MCPhoenixEngine_OLD(McEngine):
    """凤凰类 AutoCallable Monte Carlo模拟定价引擎，不支持变敲入、变敲出、变票息
            较雪球式AutoCallable，增加派息价格，少了敲出和红利票息
            每个派息（敲出）观察日，如果价格高于派息价格，派固定利息，如果发生敲出，合约提前结束；
            发生敲入后，派息方式不变，到期如果未敲出，结构为看跌空头
    """

    def __init__(self, stoch_process=None, n_path=100000, rands_method=RandsMethod.LowDiscrepancy,
                 antithetic_variate=True, ld_method=LdMethod.Sobol, seed=0, *,
                 s=None, r=None, q=None, vol=None, rands=None, s_paths=None, var_paths=None):
        """构造函数
        Args:
            stoch_process: 随机过程StochProcessBase对象
            n_path: int，MC模拟路径数
            rands_method: 生成随机数方法，RandsMethod枚举类，Pseudorandom伪随机数/LowDiscrepancy低差异序列
            antithetic_variate: bool，是否使用对立变量法
            ld_method: 若使用了低差异序列，指定低差异序列方法，LdMethod枚举类，Sobol序列/Halton序列
            seed: int，随机数种子
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process, n_path, rands_method=rands_method, antithetic_variate=antithetic_variate,
                         ld_method=ld_method, seed=seed, s=s, r=r, q=q, vol=vol, rands=rands, s_paths=s_paths,
                         var_paths=var_paths)
        # 以下为计算过程的中间变量
        self.prod = None  # Product产品对象
        self._obs_dates = None  # 根据估值日，将敲出观察日转化为List[int]，交易日期限
        self._pay_dates = None  # 根据起始日，将支付日转化为List[float]，年化自然日期限，用于计算票息
        self._n_path = None  # 随机数矩阵的路径数、期末盈亏测算=价格向量数、期末现值测算=1
        self._past_coupon = np.zeros(0, dtype=int)  # 标记过去已发生派息的数组
        # 经过估值日截断的列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
        self._barrier_out = None
        self._barrier_in = None
        self._barrier_yield = None
        self._coupon = None

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
                        不支持盈亏测算：输入spot = np.ndarray一维价格向量，t=终止日，输出payoff向量
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        obs_dates = prod.obs_dates.count_business_days(calculate_date)  # 估值日距离观察日的交易日天数
        obs_dates_list = prod.obs_dates.date_schedule
        pay_dates = prod.pay_dates.count_calendar_days(calculate_date)  # 估值日距离支付日的自然日天数

        try:  # 确定当前日期在观察日中的位置
            past_num = obs_dates_list.index(calculate_date)
        except ValueError:
            # 如果当前日期不在观察日中，找到最近的位置
            past_num = bisect.bisect_right(obs_dates_list, calculate_date)

        if spot is None:
            spot = self.process.spot()
        elif spot != self.process.spot():
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        # 当估值日就是终止日时，直接用spot计算payoff，无需随机数矩阵
        if calculate_date == prod.stop_date:
            self.s_paths = np.array([[spot, ]])  # spot是一个数值，计算期末payoff
        else:  # 随机价格路径矩阵
            self.s_paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                               t_step_per_year=prod.t_step_per_year).copy()
        self._n_path = self.s_paths.shape[1]

        if (calculate_date == prod.stop_date) and (calculate_date < prod.end_date):  # 提前敲出的参数需要特殊处理
            # 只有 obs_dates 中 = 0的观察日才会保留，之后的观察日因为敲出而消失了
            self._obs_dates = np.array([0])
            # 只取前past_num + 1个支付日, 之后的支付日因为敲出而消失了
            self._pay_dates = np.array([num / prod.annual_days.value for num in pay_dates][:past_num + 1])
            _obs_idx = obs_dates_list.index(calculate_date)
            self._barrier_out = np.array([prod.barrier_out[_obs_idx], ])
            self._barrier_in = np.array([prod.barrier_in[_obs_idx], ])
            self._barrier_yield = np.array([prod.barrier_yield[_obs_idx], ])
            self._coupon = np.array([prod.coupon[_obs_idx], ])
        else:  # 有机会持有到到期日
            self._obs_dates = np.array([num for num in obs_dates if num >= 0])  # 只取 num >= 0 (未来)的观察日
            # 包括 num < 0的支付日，从div_history取出已经发生过派息的观察日, 时间的负值可以直接输入折现因子函数
            self._pay_dates = np.array([num / prod.annual_days.value for num in pay_dates])
            # 经过估值日截断的列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
            self._barrier_out = prod.barrier_out[-len(self._obs_dates):].copy()
            self._barrier_in = prod.barrier_in[-len(self._obs_dates):].copy()
            self._barrier_yield = prod.barrier_yield[-len(self._obs_dates):].copy()
            self._coupon = prod.coupon[-len(self._obs_dates):].copy()

        # 已经发生过派息的观察日
        div_history_list = prod.div_history.date_schedule
        if div_history_list:  # 如果该列表非空
            div_history_list = list(set(div_history_list))  # 日期去重
            div_history_list.sort()  # 从小到大排序
            prod.div_history.date_schedule = div_history_list  # 更新div_history

            # 获取过去的观察日
            past_obs_dates = obs_dates_list[:past_num]
            past_coupon = np.zeros(past_num, dtype=int)  # 过去past_num个观察日的标记数组
            # 标记过去发生派息的日期
            for i, day in enumerate(past_obs_dates):
                if day in div_history_list:
                    past_coupon[i] = self._n_path
            self._past_coupon = past_coupon
        else:
            self._past_coupon = np.zeros(past_num, dtype=int)

        # 统计各个情景的payoff
        self._cal_knock_out_date()
        self._cal_yield_date()
        self._cal_knock_in_scenario()
        result = self._cal_payoff()

        if (calculate_date == prod.end_date) and isinstance(spot, np.ndarray):
            return result  # 终止日, spot是价格向量, 则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        else:
            return np.mean(result)

    def _cal_knock_out_date(self):
        """统计每条路径敲出时间"""
        # 计算每条价格路径最小敲出时间
        barrier_out = np.tile(self._barrier_out, (self._n_path, 1)).T
        barrier_out_matrix = np.tile(np.arange(self._obs_dates.shape[0]), (self._n_path, 1)).T
        barrier_out_matrix = np.where(self.s_paths[self._obs_dates, :] >= barrier_out,
                                      barrier_out_matrix.astype(int), np.inf)
        # 返回每一列的最小值，即为每条路径的敲出时间
        self.knock_out_time_idx = np.min(barrier_out_matrix, axis=0)
        # 统计哪些路径属于发生了敲出的情景(布尔索引)
        self.knock_out_scenario = (self.knock_out_time_idx != np.inf)
        # 每条路径持有时长
        hold_time_idx = self.knock_out_time_idx.copy()
        # 对于未敲出情形，持有到期，因此将索引设置为最后一个观察日的索引，如24个观察日，则指定为23
        hold_time_idx[~self.knock_out_scenario] = self._obs_dates.size - 1
        # 转成整数
        self.hold_time_idx = hold_time_idx.astype(int)

    def _cal_yield_date(self):
        """统计每条路径发生派息的时间"""
        # 统计哪些派息日，标的价格在派息线上方(即该派息日发生派息)
        barrier_yield = np.tile(self._barrier_yield, (self._n_path, 1)).T
        coupon_time_idx = (self.s_paths[self._obs_dates, :] > barrier_yield)
        # 将发生敲出之后的派息bool(不包括敲出当天)由True改为False
        self.coupon_bool = np.where(np.arange(coupon_time_idx.shape[0])[:, np.newaxis] > self.knock_out_time_idx,
                                    False, coupon_time_idx)

    def _cal_knock_in_scenario(self):
        """统计哪些路径属于敲入未敲出的情景"""
        prod = self.prod
        if prod.status == StatusType.DownTouch:  # 之前已敲入，之后未敲出的路径全部算作敲入
            self.knock_in_scenario = np.where(self.knock_out_scenario, False, True)
        else:
            if prod.in_obs_type == ExerciseType.American:  # 每日观察敲入(凤凰)
                # 排除发生了敲出的路径，统计哪些路径属于敲入未敲出
                knock_in_level = np.array(self._barrier_in).repeat(
                    np.diff(np.append(np.zeros((1,)), self._obs_dates)).astype(int))
                knock_in_level = np.append(self._barrier_in[0], knock_in_level)
                knock_in_level = np.tile(knock_in_level, (self._n_path, 1)).T
                knock_in_time_idx = (self.s_paths <= knock_in_level)
                # 统计哪些路径属于敲入未敲出
                knock_in_bool = np.any(knock_in_time_idx, axis=0)
            elif prod.in_obs_type == ExerciseType.European:  # 到期观察敲入(DCN/FCN)
                knock_in_bool = (self.s_paths[-1] <= self._barrier_in[-1])
            else:
                raise ValueError(f"不支持的敲入观察类型{prod.in_obs_type}，仅支持美式(凤凰)/欧式(DCN/FCN)")
            # 将发生敲出之后的敲入由True改为False
            self.knock_in_scenario = np.where(self.knock_out_scenario, False, knock_in_bool)

    def _cal_payoff(self):
        """统计每条路径的收益加总"""
        # 不同派息日的票息分别计算贴现因子，得到派息的现值
        discount_factor = np.empty(self._pay_dates.size)
        for i, pay_d in enumerate(self._pay_dates):
            discount_factor[i] = self.process.interest.disc_factor(pay_d)  # 每月的派息会复利
            # discount_factor[i] = self.process.interest.disc_factor(self._pay_dates[-1])  # 每月的派息不复利
        discounted_coupon = self._coupon * self.prod.s0 * discount_factor

        # payoff汇总
        payoff = 0
        # 1.未敲入的部分（敲出/未敲入未敲出），到期还本
        payoff += np.sum(discount_factor[self.hold_time_idx[~self.knock_in_scenario]]
                         ) * self.prod.margin_lvl * self.prod.s0

        # 2.派息payoff
        paths_sum_coupon_times = np.sum(self.coupon_bool, axis=1)
        # 拼接过去已发生的派息次数数组
        full_coupon = np.concatenate([self._past_coupon, paths_sum_coupon_times])
        # assert len(full_coupon) != len(discounted_coupon), "票息数组与折现因子数组长度应该相等"
        payoff += np.sum(full_coupon * discounted_coupon)

        # 3.敲入，承担跌幅损失
        s_vec = self.s_paths[-1, self.knock_in_scenario].copy()
        payoff += (np.sum(self.prod.margin_lvl * self.prod.s0 + self.prod.parti_in *
                          np.maximum(
                              np.minimum(s_vec - self.prod.strike_upper, 0),
                              self.prod.strike_lower - self.prod.strike_upper))
                   ) * discount_factor[-1]
        payoff /= self._n_path
        return payoff


class MCPhoenixEngine(McEngine):
    """凤凰类 AutoCallable Monte Carlo模拟定价引擎，不支持变敲入、变敲出、变票息
            较雪球式AutoCallable，增加派息价格，少了敲出和红利票息
            每个派息（敲出）观察日，如果价格高于派息价格，派固定利息，如果发生敲出，合约提前结束；
            发生敲入后，派息方式不变，到期如果未敲出，结构为看跌空头
    """

    def __init__(self, stoch_process=None, n_path=100000, rands_method=RandsMethod.LowDiscrepancy,
                 antithetic_variate=True, ld_method=LdMethod.Sobol, seed=0, *,
                 s=None, r=None, q=None, vol=None, rands=None, s_paths=None, var_paths=None):
        """构造函数
        Args:
            stoch_process: 随机过程StochProcessBase对象
            n_path: int，MC模拟路径数
            rands_method: 生成随机数方法，RandsMethod枚举类，Pseudorandom伪随机数/LowDiscrepancy低差异序列
            antithetic_variate: bool，是否使用对立变量法
            ld_method: 若使用了低差异序列，指定低差异序列方法，LdMethod枚举类，Sobol序列/Halton序列
            seed: int，随机数种子
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process, n_path, rands_method=rands_method, antithetic_variate=antithetic_variate,
                         ld_method=ld_method, seed=seed, s=s, r=r, q=q, vol=vol, rands=rands, s_paths=s_paths,
                         var_paths=var_paths)
        # 以下为计算过程的中间变量
        self.prod = None  # Product产品对象
        self._obs_dates = None  # 根据估值日，将敲出观察日转化为List[int]，交易日期限
        self._div_obs_dates = None  # 根据估值日，将派息观察日转化为List[int]，交易日期限
        self._pay_dates = None  # 根据起始日，将支付日转化为List[float]，年化自然日期限，用于计算票息
        self._n_path = None  # 随机数矩阵的路径数、期末盈亏测算=价格向量数、期末现值测算=1
        self._past_coupon = np.zeros(0, dtype=int)  # 标记过去已发生派息的数组
        # 经过估值日截断的列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
        self._barrier_out = None
        self._barrier_in = None
        self._barrier_yield = None
        self._coupon = None

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
                        不支持盈亏测算：输入spot = np.ndarray一维价格向量，t=终止日，输出payoff向量
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)

        obs_dates = prod.obs_dates.count_business_days(calculate_date)  # 估值日距离敲出观察日的交易日天数
        obs_dates_list = prod.obs_dates.date_schedule
        div_obs_dates = prod.div_obs_dates.count_business_days(calculate_date)  # 估值日距离派息观察日的交易日天数
        div_obs_dates_list = prod.div_obs_dates.date_schedule
        pay_dates = prod.pay_dates.count_calendar_days(calculate_date)  # 估值日距离支付日的自然日天数
        digital_num = prod.lock_term - 1

        try:  # 确定当前日期在观察日中的位置
            # past_num = obs_dates_list.index(calculate_date)
            past_div_num = div_obs_dates_list.index(calculate_date)
        except ValueError:
            # 如果当前日期不在观察日中，找到最近的位置
            # past_num = bisect.bisect_right(obs_dates_list, calculate_date)
            past_div_num = bisect.bisect_right(div_obs_dates_list, calculate_date)

        if spot is None:
            spot = self.process.spot()

        # 当估值日就是终止日时，直接用spot计算payoff，无需随机数矩阵
        if calculate_date == prod.stop_date:
            if isinstance(spot, (list, np.ndarray)):  # spot是一个np.ndarray一维价格向量，返回盈亏测算向量，即每个价格对应的现值，而不是平均值
                self.s_paths = np.array([spot, ])
            else:  # spot是一个数值，计算期末payoff
                self.s_paths = np.array([[spot, ]])
        else:  # 随机价格路径矩阵
            if spot != self.process.spot():
                self.reset_paths_flag()  # 重置路径标志位，重新生成路径
            self.s_paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                               t_step_per_year=prod.t_step_per_year).copy()
        self._n_path = self.s_paths.shape[1]

        if (calculate_date == prod.stop_date) and (calculate_date < prod.end_date):  # 提前敲出的参数需要特殊处理
            # 只有 obs_dates 中 = 0的观察日才会保留，之后的观察日因为敲出而消失了
            self._obs_dates = np.array([0])
            self._div_obs_dates = np.array([0]) if 0 in div_obs_dates else np.array([])
            # 只取前past_div_num个支付日, 之后的支付日因为敲出而消失了
            self._pay_dates = np.array([num / prod.annual_days.value for num in pay_dates][:past_div_num + 1])
            _obs_idx = obs_dates_list.index(calculate_date)
            self._barrier_out = np.array([prod.barrier_out[_obs_idx], ])
            self._barrier_in = np.array([prod.barrier_in[_obs_idx + digital_num], ])
            self._barrier_yield = np.array([prod.barrier_yield[_obs_idx + digital_num], ])
            self._coupon = prod.coupon[:past_div_num + 1]
        else:  # 有机会持有到到期日
            self._obs_dates = np.array([num for num in obs_dates if num >= 0])  # 只取 num >= 0 (未来)的敲出观察日
            # 包括 num < 0的支付日，从div_history取出已经发生过派息的观察日, 时间的负值可以直接输入折现因子函数
            self._div_obs_dates = np.array([num for num in div_obs_dates if num >= 0])  # 只取 num >= 0 (未来)的派息观察日
            self._pay_dates = np.array([num / prod.annual_days.value for num in pay_dates])
            # 经过估值日截断的列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
            self._barrier_out = prod.barrier_out[-len(self._obs_dates):].copy()
            self._barrier_in = prod.barrier_in[-len(self._div_obs_dates):].copy()
            self._barrier_yield = prod.barrier_yield[-len(self._div_obs_dates):].copy()
            self._coupon = prod.coupon[-len(self._pay_dates):].copy()

        # 已经发生过派息的观察日
        div_history_list = prod.div_history.date_schedule
        if div_history_list:  # 如果该列表非空
            div_history_list = list(set(div_history_list))  # 日期去重
            div_history_list.sort()  # 从小到大排序
            prod.div_history.date_schedule = div_history_list  # 更新div_history

            # 获取过去的派息观察日
            past_div_obs_dates = div_obs_dates_list[:past_div_num]
            past_coupon = np.zeros(past_div_num, dtype=int)  # 过去past_num个观察日的标记数组
            # 标记过去发生派息的日期
            for i, day in enumerate(past_div_obs_dates):
                if day in div_history_list:
                    past_coupon[i] = True
            self._past_coupon = past_coupon
        else:
            self._past_coupon = np.zeros(past_div_num, dtype=int)

        # 统计各个情景的payoff
        self._cal_knock_out_date()
        self._cal_yield_date()
        self._cal_knock_in_scenario()
        result = self._cal_payoff()

        if (calculate_date == prod.end_date) and isinstance(spot, np.ndarray):
            return result  # 终止日, spot是价格向量, 则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        else:
            return np.mean(result)

    def _cal_knock_out_date(self):
        """统计每条路径敲出时间"""
        # 计算每条价格路径最小敲出时间
        barrier_out = np.tile(self._barrier_out, (self._n_path, 1)).T
        barrier_out_matrix = np.tile(np.arange(self._obs_dates.shape[0]), (self._n_path, 1)).T
        barrier_out_matrix = np.where(self.s_paths[self._obs_dates, :] >= barrier_out,
                                      barrier_out_matrix.astype(int), np.inf)
        # 返回每一列的最小值，即为每条路径的敲出时间
        self.knock_out_time_idx = np.min(barrier_out_matrix, axis=0)
        # 统计哪些路径属于发生了敲出的情景(布尔索引)
        self.knock_out_scenario = (self.knock_out_time_idx != np.inf)
        # 每条路径持有时长, 用于计算返还保证金的折现因子
        hold_time_idx = self.knock_out_time_idx.copy()
        # 对于未敲出情形，持有到期，因此将索引设置为最后一个观察日的索引，如24个观察日，则指定为23
        hold_time_idx[~self.knock_out_scenario] = self._obs_dates.size - 1
        # 由于敲出观察日会根据估值日的位置被截断，因此，应该补齐敲出观察日比派息观察日少的数量，得到的持有时长索引才能与派息支付日的索引对齐
        self.hold_time_idx = hold_time_idx.astype(int) + len(self._pay_dates) - len(self._obs_dates)

    def _cal_yield_date(self):
        """统计每条路径发生派息的时间"""
        # 统计哪些派息日，标的价格在派息线上方(即该派息日发生派息)
        barrier_yield = np.tile(self._barrier_yield, (self._n_path, 1)).T
        coupon_time_idx = (self.s_paths[self._div_obs_dates, :] > barrier_yield)
        # 利用每条路径持有时长的索引数组, 将发生敲出之后的派息bool(不包括敲出当天)由True改为False
        self.coupon_bool = np.where(np.arange(coupon_time_idx.shape[0])[:, np.newaxis] > self.hold_time_idx,
                                    False, coupon_time_idx)

    def _cal_knock_in_scenario(self):
        """统计哪些路径属于敲入未敲出的情景"""
        prod = self.prod
        if prod.status == StatusType.UpTouch:  # 已敲出
            self.knock_in_scenario = np.zeros(self._n_path)
        elif prod.status == StatusType.DownTouch:  # 之前已敲入，之后未敲出的路径全部算作敲入
            self.knock_in_scenario = np.where(self.knock_out_scenario, False, True)
        else:
            if prod.in_obs_type == ExerciseType.American:  # 每日观察敲入(凤凰)
                # 排除发生了敲出的路径，统计哪些路径属于敲入未敲出
                knock_in_level = np.array(self._barrier_in).repeat(
                    np.diff(np.append(np.zeros((1,)), self._div_obs_dates)).astype(int))
                knock_in_level = np.append(self._barrier_in[0], knock_in_level)
                knock_in_level = np.tile(knock_in_level, (self._n_path, 1)).T
                knock_in_time_idx = (self.s_paths <= knock_in_level)
                # 统计哪些路径属于敲入未敲出
                knock_in_bool = np.any(knock_in_time_idx, axis=0)
            elif prod.in_obs_type == ExerciseType.European:  # 到期观察敲入(DCN/FCN)
                knock_in_bool = (self.s_paths[-1] <= self._barrier_in[-1])
            else:
                raise ValueError(f"不支持的敲入观察类型{prod.in_obs_type}，仅支持美式(凤凰)/欧式(DCN/FCN)")
            # 将发生敲出之后的敲入由True改为False
            self.knock_in_scenario = np.where(self.knock_out_scenario, False, knock_in_bool)

    def _cal_payoff(self):
        """统计每条路径的收益加总"""
        # 不同派息日的票息分别计算贴现因子，得到派息的现值
        discount_factor = np.empty(self._pay_dates.size)
        for i, pay_d in enumerate(self._pay_dates):
            discount_factor[i] = self.process.interest.disc_factor(pay_d)  # 每月的派息会复利
        discounted_coupon = self._coupon * self.prod.s0 * discount_factor

        # payoff汇总
        # 1. 归还保证金(敲入亏损在3.另行计算)
        payoff1 = self.prod.margin_lvl * self.prod.s0 * discount_factor[self.hold_time_idx]

        # 2. 派息payoff
        # 过去已发生的派息
        past_value = np.dot(self._past_coupon, discounted_coupon[:len(self._past_coupon)])  # 向量内积
        # 模拟路径发生的派息
        future_value = discounted_coupon[len(self._past_coupon):] @ self.coupon_bool  # 矩阵乘法
        payoff2 = past_value + future_value

        # 3. 敲入的路径，承担跌幅损失
        payoff3 = np.zeros_like(self.s_paths[-1])
        if np.any(self.knock_in_scenario):
            s_knock_in = self.s_paths[-1][self.knock_in_scenario]
            payoff_knock_in = self.prod.parti_in * discount_factor[-1] * np.minimum(
                np.maximum(s_knock_in, self.prod.strike_lower) - self.prod.strike_upper, 0)
            payoff3[self.knock_in_scenario] = payoff_knock_in

        payoff = payoff1 + payoff2 + payoff3
        return payoff



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.utilities.enums import InOut, ExerciseType, PaymentType
from pricelib.common.date_handler import global_evaluation_date
from pricelib.common.pricing_engine_base import McEngine


class MCDoubleBarrierEngine(McEngine):
    """双边障碍期权 Monte Carlo 模拟定价引擎
    支持欧式观察(仅到期观察)/美式观察(整个有效期观察)；只支持离散观察(默认为每日观察)；
    支持现金返还；敲入现金返还为到期支付；敲出现金返还支持 立即支付/到期支付"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        if spot is None:
            spot = self.process.spot()
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value  # 从估值日到到期日的自然日时间（年）
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date,
                                                                            prod.end_date)  # 从估值日到到期日的交易日时间（天）
        _elapsed_days = prod.trade_calendar.business_days_between(prod.start_date, calculate_date)  # 从起始日到估值日的交易日时间（天）

        # 当估值日就是终止日时，直接用spot计算payoff，无需随机数矩阵
        if (calculate_date == prod.end_date) and isinstance(spot, np.ndarray):
            # 如果spot是一个np.ndarray一维价格向量，返回盈亏测算向量，即每个价格对应的现值，而不是平均值
            _tenor = (prod.end_date - prod.start_date).days / prod.annual_days.value  # 从起始日到终止日的自然日时间（年）
            _floor_yield = self._preprocess_floor_yield(prod, _tenor)
            value = self.calc_maturity_payoff(prod, np.array([spot, ]), _tenor, _floor_yield)
            return value
        elif calculate_date == prod.stop_date:  # spot是一个数值，计算到期日payoff
            _tenor = _tenor = (prod.stop_date - prod.start_date).days / prod.annual_days.value  # 从起始日到终止日的自然日时间（年）
            _floor_yield = self._preprocess_floor_yield(prod, _tenor)
            value = self.calc_maturity_payoff(prod, np.array([spot, ]), _tenor, _floor_yield)[0]
            return value
        else:
            if spot != self.process.spot():
                self.reset_paths_flag()  # 重置路径标志位，重新生成路径
            _tenor = (prod.end_date - prod.start_date).days / prod.annual_days.value  # 从起始日到终止日的自然日时间（年）
            _floor_yield = self._preprocess_floor_yield(prod, _tenor)
            paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()

        r = self.process.interest(_maturity)
        # 美式观察：
        if prod.obs_type == ExerciseType.American:
            payoff = prod.callput.value * (paths[-1] - prod.strike) * prod.parti  # 香草的payoff
            if prod.discrete_obs_interval is None:  # 每日观察
                obs_points = np.arange(0, _maturity_business_days + 1, 1)
            else:  # 均匀离散观察
                dt_step = prod.discrete_obs_interval * prod.t_step_per_year
                obs_points = np.flip(np.round(np.arange(_maturity_business_days, 0, -dt_step)).astype(int))
                obs_points[obs_points > _maturity_business_days] = _maturity_business_days  # 防止闰年导致的下标越界

            if prod.inout == InOut.In:  # 美式双边敲入一定是到期支付
                knock_in_bool = np.any((paths[obs_points] <= prod.bound[0]) | (paths[obs_points] >= prod.bound[1]),
                                       axis=0)
                _rebate = self._preprocess_extra_returns(prod, _tenor, [prod.rebate[0], ])
                value = np.where(knock_in_bool, np.where(payoff < 0, 0, payoff) + _floor_yield, _rebate)
                value = np.mean(value) * math.exp(-r * _maturity)
            elif prod.inout == InOut.Out:  # 美式双边敲出
                step_index = np.tile(obs_points, (self.n_path, 1)).T.astype(int)
                hit_lower = np.min(
                    np.where(paths[1:] <= prod.bound[0], step_index.astype(int), _maturity_business_days + 1),
                    axis=0)
                hit_upper = np.min(
                    np.where(paths[1:] >= prod.bound[1], step_index.astype(int), _maturity_business_days + 1),
                    axis=0)
                if prod.payment_type == PaymentType.Hit:  # 发生敲出时，立即支付现金返还
                    _rebate_lower = self._preprocess_extra_returns(prod,
                                                                   (_elapsed_days + hit_lower[hit_lower < hit_upper]
                                                                    ) / prod.t_step_per_year, [prod.rebate[0], ])
                    _rebate_upper = self._preprocess_extra_returns(prod,
                                                                   (_elapsed_days + hit_upper[hit_lower > hit_upper]
                                                                    ) / prod.t_step_per_year, [prod.rebate[1], ])
                    value = np.mean(
                        np.where(hit_lower < hit_upper, _rebate_lower * np.exp(-r * hit_lower / prod.t_step_per_year),
                                 np.where(hit_lower > hit_upper,
                                          _rebate_upper * np.exp(-r * hit_upper / prod.t_step_per_year),
                                          (np.where(payoff < 0, 0, payoff) + _floor_yield) * math.exp(-r * _maturity))))
                elif prod.payment_type == PaymentType.Expire:  # 发生敲出时，到期支付现金返还
                    _rebate = self._preprocess_extra_returns(prod, _tenor, prod.rebate)
                    value = np.mean(np.where(hit_lower < hit_upper, _rebate[0],
                                             np.where(hit_lower > hit_upper, _rebate[1],
                                                      np.where(payoff < 0, 0, payoff) + _floor_yield))
                                    ) * math.exp(-r * _maturity)
                else:
                    raise ValueError("PaymentType must be Hit or Expire")
            else:
                raise ValueError("InOut must be In or Out")
        # 欧式（到期观察）：
        elif prod.obs_type == ExerciseType.European:
            value = self.calc_maturity_payoff(prod, paths[-1], _tenor, _floor_yield)
            value = np.mean(value) * math.exp(-r * _maturity)
        else:
            raise ValueError("ExerciseType must be American or European")
        return value

    def calc_maturity_payoff(self, prod, s_vec, _tenor, _floor_yield=0.):
        """初始化终止时的期权价值，在障碍价格内侧，默认设定未敲入，未敲出
        Args:
            prod: Product产品对象
            s_vec: np.ndarray, 网格的价格格点
            _tenor: float, 从起始日到终止日的自然日时间（年）
            _floor_yield: float，处理后的保底收益绝对数值
        Returns:
            value: np.ndarray, 期末价值向量
        """
        payoff = prod.callput.value * (s_vec - prod.strike) * prod.parti
        if prod.inout == InOut.Out:
            _rebate = self._preprocess_extra_returns(prod, _tenor, prod.rebate)
            value = np.where(s_vec <= prod.bound[0], _rebate[0],
                             np.where(s_vec >= prod.bound[1], _rebate[1],
                                      np.where(payoff < 0, 0, payoff) + _floor_yield))
        elif prod.inout == InOut.In:
            _rebate = self._preprocess_extra_returns(prod, _tenor, [prod.rebate[0], ])
            value = np.where((s_vec > prod.bound[0]) & (s_vec < prod.bound[1]), _rebate,
                             np.where(payoff < 0, 0, payoff) + _floor_yield)
        else:
            raise ValueError("InOut must be In or Out")
        return value

    @staticmethod
    def _preprocess_extra_returns(prod, period, rebate_list):
        """预处理可能存在的年化现金返还rebate
        Args:
            prod: Product产品对象
            period: float 或 np.ndarray, 从起始到期权终止的时间（年）
                    对欧式障碍、美式敲入而言是整个存续期；对美式敲出而言，期权可能会因为敲出提前终止，此时period是起始日到敲出日的时间
            rebate_list: List[float], 现金返还率或绝对数值
        Returns:
            _rebate: float，处理后的现金返还绝对数值
        """
        if prod.rebate_annual:  # 年化百分比
            _rebate = []
            for r in rebate_list:
                _rebate.append(r * period * prod.s0)
        else:
            _rebate = rebate_list  # 绝对数值
        return _rebate

    @staticmethod
    def _preprocess_floor_yield(prod, period):
        """预处理可能存在的年化保底收益floor_yield
        Args:
            prod: Product产品对象
            period: float 或 np.ndarray, 从起始到期权终止的时间（年）
                    对欧式障碍、美式敲入而言是整个存续期；对美式敲出而言，期权可能会因为敲出提前终止，此时period是起始日到敲出日的时间
        Returns:
            _floor_yield: float，处理后的保底收益绝对数值
        """
        if prod.floor_yield != 0:  # 保底收益率
            _floor_yield = prod.floor_yield * period * prod.s0
        else:
            _floor_yield = 0.0
        return _floor_yield


#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.pricing_engine_base import QuadEngine
from pricelib.common.date_handler import global_evaluation_date


class QuadPortfolioEngine(QuadEngine):
    """香草组合积分法定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None, calc_greeks=False, s_step=0.001):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
            calc_greeks: bool，是否计算delta和gamma，默认为False，即只计算理论估值
            s_step: float，价格网格步长（百分比）
        Returns:
            若只计算pv，返回float，现值 pv
            若计算pv、delta和gamma，返回np.ndarray [down_price, pv, up_price] ，其中 down_price 和 up_price 是 s±Δs 的理论估值
        """
        calculate_date = global_evaluation_date() if t is None else t
        tau = (prod.end_date - calculate_date).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()

        if tau == 0:  # 如果估值日是到期日
            if calc_greeks:  # 计算理论估值、delta和gamma
                value = self._init_terminal_condition(prod, np.array([1 - s_step, 1, 1 + s_step]) * spot)
            else:  # 只计算理论估值, 直接计算终止条件
                value = self._init_terminal_condition(prod, np.array([spot]))[0]
            return value

        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)
        self._check_method_params()

        # 设定期末价值
        self.init_grid(spot, vol, tau)
        v_vec = np.zeros(self.ln_s_vec.shape)
        for vanilla, position in prod.vanilla_list:
            v_vec += np.maximum(vanilla.callput.value * (np.exp(self.ln_s_vec) - vanilla.strike), 0) * position
        # 设置积分法engine参数
        self.set_quad_params(r=r, q=q, vol=vol)
        # 回溯计算
        if calc_greeks:  # 计算理论估值、delta和gamma
            up_price = self.fft_step_backward(np.array([np.log(spot * (1 + s_step))]), self.ln_s_vec, v_vec, tau)[0]
            pv = self.fft_step_backward(np.array([np.log(spot)]), self.ln_s_vec, v_vec, tau)[0]
            down_price = self.fft_step_backward(np.array([np.log(spot * (1 - s_step))]), self.ln_s_vec, v_vec, tau)[0]
            return np.array([down_price, pv, up_price])
        else:  # 只计算理论估值
            value = self.fft_step_backward(np.log(np.array([spot])), self.ln_s_vec, v_vec, tau)[0]
        return value

    @staticmethod
    def _init_terminal_condition(prod, s_vec):
        """初始化终止时的期权价值，在障碍价格内侧，默认设定未敲入，未敲出
        Args:
            prod: Product产品对象
            s_vec: np.ndarray, 网格的价格格点
        Returns:
            payoff: np.ndarray, 期末价值向量
        """
        value = np.zeros(len(s_vec))
        for vanilla, position in prod.vanilla_list:
            value += np.maximum(vanilla.callput.value * (s_vec - vanilla.strike), 0) * position
        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import bisect
from contextlib import suppress
import datetime
from typing import Callable
from functools import partial
import numpy as np
from pricelib.common.utilities.enums import StatusType, ExerciseType, EngineType
from pricelib.common.date_handler import (global_evaluation_date, Schedule, CN_CALENDAR, AnnualDays)
from pricelib.common.utilities.patterns import Observer, global_auto_update_status
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase

from .european_snowball import EuropeanSnowball
from .floored_snowball import FlooredSnowball
from pricelib.products.digital.digital_option import DigitalOption
from pricelib.pricing_engines.mc_engines import MCPhoenixEngine
from pricelib.pricing_engines.fdm_engines import FdmPhoenixEngine, FdmDigitalEngine, FdmSnowBallEngine
from pricelib.pricing_engines.integral_engines import (
    QuadPhoenixEngine, QuadDigitalEngine, QuadEuroSnowEngine, QuadSnowballEngine)

PHOENIX_ENGINE_DICT = {
    EngineType.QuadEngine: {
        'DigitalOption': QuadDigitalEngine,
        'EuropeanSnowball': QuadEuroSnowEngine,
        'FlooredSnowball': QuadSnowballEngine},
    EngineType.PdeEngine: {
        'DigitalOption': FdmDigitalEngine,
        'EuropeanSnowball': FdmSnowBallEngine,
        'FlooredSnowball': FdmSnowBallEngine},
}


class PhoenixBase(OptionBase, Observer):
    """FCN/DCN/凤凰结构定价基类
        凤凰类 AutoCallable
            较雪球式AutoCallable，增加派息价格，少了敲出和红利票息
            每个派息（敲出）观察日，如果价格高于派息价格，派固定利息，如果发生敲出，合约提前结束；
            发生敲入后，派息方式不变，到期如果未敲出，结构为看跌空头或熊市价差空头
        FCN(Fixed Coupon Note, 固定票息票据)/DCN(Dynamic Coupon Note, 动态票息票据)/Phoenix凤凰票据
            FCN：无派息价格（派息价格为0），每月固定派息；到期日观察是否敲入
            DCN：有派息价格，每个观察日如果价格高于派息价格，派固定利息；到期日观察是否敲入
            Phoenix：有派息价格，每个观察日如果价格高于派息价格，派固定利息；每日观察是否敲入
    """

    def __init__(self, s0,
                 barrier_out=None, barrier_yield=None, coupon=None, barrier_in=None, parti_in=1, margin_lvl=1,
                 strike_upper=None, strike_lower=0, obs_dates=None, pay_dates=None, div_obs_dates=None, lock_term=None,
                 in_obs_type=ExerciseType.American, status=StatusType.NoTouch, div_history=None, engine=None,
                 maturity=None, start_date=None, end_date=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365,
                 t_step_per_year=243, stoch_process=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """构造函数
        Args:
            s0: float，标的初始价格
            barrier_out: List[float]，敲出障碍价，绝对值/百分比，长度需要与敲出观察日数一致
                         若不是常数，为阶梯式凤凰（降敲）
            barrier_yield: List[float]，派息边界价格，绝对值/百分比，长度需要与派息观察日数一致
                           凤凰和DCN，barrier_yield一般等于敲入价格；FCN每月固定派息，barrier_yield等于0
            coupon: List[float]，派息票息，百分比，非年化
                    数组list，长度需要与派息观察日数一致，为阶梯式浮动票息（降息或双降）凤凰
            barrier_in: List[float]，敲入障碍价，绝对值/百分比
                    若不是常数，为变敲入型雪球，每个敲出观察日对应一个敲入价，在这个敲出观察日和上个敲出观察日之间，敲入价为该敲入价
                    对于FCN，进到期日观察是否敲入，到期日前的敲入价都是0
            parti_in: float，敲入后参与率，限损
            margin_lvl: float，保证金比例，默认为1，即无杠杆
            strike_upper: float，敲入发生后(熊市价差)的高行权价，即OTM行权价(<s0)，普通凤凰敲入发生后的高行权价为期初价格s0
            strike_lower: float，敲入发生后(熊市价差)的低行权价，即保底边界，越高提供的保护越多，值为0时不保底
            lock_term: int，锁定期，单位为月，锁定期内不触发敲出
            status: 敲入敲出状态，StatusType枚举类，默认为NoTouch，已敲入未敲出为DownTouch，已敲出为UpTouch
            div_history: List[datetime.date]，历史派息列表，记录已经发生派息的派息观察日div_obs_dates
            in_obs_type: 敲入观察类型，ExerciseType枚举类，默认为American每日观察；European为仅到期日观察是否敲入
        时间参数: 要么输入年化期限，要么输入起始日和到期日；敲出观察日和票息支付日可缺省
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_dates: List[datetime.date]，敲出观察日，可缺省，缺省时会自动生成每月观察的敲出日期序列(已根据节假日调整)
            pay_dates: List[datetime.date]，票息支付日，可缺省，缺省时会自动生成每月观察的派息日期序列(已根据节假日调整)
            div_obs_dates: List[datetime.date]，派息观察日，可缺省，缺省时会自动生成每月观察的派息观察日期序列（已根据节假日调整）
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            engine: 定价引擎，PricingEngine类
                    蒙特卡洛: MCPhoenixEngine
                    PDE: FdmPhoenixEngine
                    积分法: QuadPhoenixEngine
            engine_method: str, 定价引擎种类。若未输入engine，可以用engine_method快捷生成指定定价引擎。可选MC/MonteCarlo、PDE/FDM
            若未提供引擎engine、也未指定引擎种类engine_method的情况下，如果提供了 随机过程 或 标的价格、无风险利率、分红/融券率、波动率，
            则默认使用Quad定价引擎 QuadPhoenixEngine
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.s0 = s0  # 标的初始价格
        self.trade_calendar = trade_calendar  # 交易日历
        self.annual_days = annual_days
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if self.trade_calendar.isholiday(self.end_date):  # 如果到期日不是交易日，向后顺延一个交易日
            self.end_date = self.trade_calendar.advance(self.end_date, datetime.timedelta(days=1))
        if maturity is None:
            self.maturity = (end_date - start_date).days / annual_days.value
        else:
            self.maturity = maturity

        # 敲出观察日
        if obs_dates is None:
            self.obs_dates = Schedule(trade_calendar=trade_calendar, start=self.start_date,
                                      end=self.end_date, freq='m', lock_term=lock_term)  # 有锁定期
            self.end_date = self.obs_dates.date_schedule[-1]  # 修正整个产品的到期时间
        else:
            self.obs_dates = Schedule(trade_calendar=trade_calendar, date_schedule=obs_dates)
        # 派息观察日
        if div_obs_dates is None:
            self.div_obs_dates = Schedule(trade_calendar=trade_calendar, start=self.start_date,
                                          end=self.end_date, freq='m', lock_term=1)  # 无锁定期
        else:
            self.div_obs_dates = Schedule(trade_calendar=trade_calendar, date_schedule=div_obs_dates)
        # 支付日
        if pay_dates is None:
            self.pay_dates = self.div_obs_dates  # 默认等于派息观察日
        else:
            self.pay_dates = Schedule(trade_calendar=trade_calendar, date_schedule=pay_dates)
        assert len(self.div_obs_dates.date_schedule) == len(self.pay_dates.date_schedule), "派息观察日与支付日期列表长度应该相等!"
        self.t_step_per_year = t_step_per_year
        if lock_term is None:
            self.lock_term = len(self.div_obs_dates.date_schedule) - len(self.obs_dates.date_schedule) + 1
        else:
            self.lock_term = lock_term  # 锁定期
            assert lock_term - 1 == len(self.div_obs_dates.date_schedule) - len(self.obs_dates.date_schedule), "敲出观察日与派息观察日的个数之差与锁定期不匹配!"
        # 敲出线
        if isinstance(barrier_out, (int, float, np.int32, np.float64)):
            self.barrier_out = (np.ones(len(self.obs_dates.date_schedule)) * barrier_out).astype(float)  # 敲出线与敲出观察日等长
        elif isinstance(barrier_out, (list, np.ndarray)):
            self.barrier_out = np.array(barrier_out)
        else:
            raise ValueError(f'敲出价类型为{type(barrier_out)}，仅支持int/float/list/np.ndarray，请检查')
        # 敲入线
        if isinstance(barrier_in, (int, float, np.int32, np.float64)):
            self.barrier_in = np.ones(len(self.div_obs_dates.date_schedule)) * barrier_in  # 敲入线与派息观察日等长
        elif isinstance(barrier_in, (list, np.ndarray)):
            self.barrier_in = np.array(barrier_in)
        else:
            raise ValueError(f'敲出价类型为{type(barrier_out)}，仅支持int/float/list/np.ndarray，请检查')
        # 派息边界价格
        if barrier_yield is None:
            self.barrier_yield = self.barrier_in  # 派息线默认等于敲入线
        else:
            if isinstance(barrier_yield, (int, float, np.int32, np.float64)):
                self.barrier_yield = np.ones(len(self.div_obs_dates.date_schedule)) * barrier_yield  # 派息线与派息观察日等长
            elif isinstance(barrier_yield, (list, np.ndarray)):
                self.barrier_yield = np.array(barrier_yield)
            else:
                raise ValueError(f'派息价类型为{type(barrier_yield)}，仅支持int/float/list/np.ndarray，请检查')
        # 票息(绝对收益，默认为月度)
        if isinstance(coupon, (int, float, np.int32, np.float64)):
            self.coupon = np.ones(len(self.pay_dates.date_schedule)) * coupon  # 票息数组与派息支付日等长
        elif isinstance(coupon, (list, np.ndarray)):
            self.coupon = np.array(coupon)
        else:
            raise ValueError(f'派息比例类型为{type(coupon)}，仅支持int/float/list/np.ndarray，请检查')
        self.parti_in = parti_in  # 下跌参与率
        self.margin_lvl = margin_lvl  # 预付金比例
        self.strike_upper = s0 if strike_upper is None else strike_upper  # 敲入后高执行价
        self.strike_lower = strike_lower  # 低行权价
        self.status = status
        # 派息历史，记录已经发生派息的派息观察日div_obs_dates
        if div_history is not None:
            self.div_history = Schedule(trade_calendar=trade_calendar, start=self.start_date, end=self.end_date,
                                        freq='m', lock_term=1, date_schedule=div_history)
        else:
            self.div_history = Schedule(trade_calendar=trade_calendar, start=self.start_date, end=self.end_date,
                                        freq='m', lock_term=1, date_schedule=[])
        self.in_obs_type = in_obs_type  # 敲入观察为欧式，到期观察敲入；美式，每日观察敲入
        self.stop_date = self.end_date  # 停止日期，默认等于到期日。发生敲出事件，停止日期更新为敲出日期

        # 凤凰/DCN拆分成的25个子期权self.sub_products, 用self.init_sub_products()方法初始化
        self.sub_products = None
        self.init_sub_products()
        # 反算票息
        self.calc_coupon = partial(self.calc_param, reset_param_method=self.reset_coupon)

        if engine is not None:
            self.set_pricing_engine(engine)
        else:
            if engine_method is None:
                default_engine = QuadPhoenixEngine  # 默认定价引擎
            elif engine_method in ["PDE", "FDM"]:
                default_engine = FdmPhoenixEngine
            elif engine_method in ["MC", "MonteCarlo"]:
                default_engine = MCPhoenixEngine
            elif engine_method == "QUAD":
                default_engine = QuadPhoenixEngine
            else:
                raise NotImplementedError(f"不支持的engine_method{engine_method}, 凤凰、FCN、DCN仅支持MC/PDE/QUAD")
            if stoch_process is not None:
                self.set_pricing_engine(default_engine(stoch_process))
            elif s is not None and r is not None and q is not None and vol is not None:
                self.set_pricing_engine(default_engine(s=s, r=r, q=q, vol=vol))

    def set_pricing_engine(self, engine):
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

        self_engine_type = self.engine.engine_type
        if self_engine_type in [EngineType.PdeEngine, EngineType.QuadEngine]:
            # PDE-FDM有限差分法 或 QUAD积分法，需要为子产品设置对应的定价引擎
            for sub_product in self.sub_products:
                sub_product_class = sub_product.__class__.__name__
                sub_engine_type = PHOENIX_ENGINE_DICT[self_engine_type][sub_product_class]
                sub_engine = sub_engine_type(self.engine.process)
                sub_product.set_pricing_engine(sub_engine)
        return

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新自动赎回结构是否已经敲入或敲出; 以及记录是否发生派息"""
        if global_auto_update_status():
            calculate_date = global_evaluation_date()
            if calculate_date < self.start_date or calculate_date > self.stop_date:
                return  # 如果在起始日前或终止日之后，状态维持不变
            if observable == self.engine.process.spot:
                if calculate_date in self.div_obs_dates.date_schedule:  # 如果估值日是派息观察日
                    div_obs_idx = self.div_obs_dates.date_schedule.index(calculate_date)
                    if observable.data >= self.barrier_yield[div_obs_idx]:  # 如果标的价格大于等于派息价，记录发生派息的观察日
                        self.div_history.date_schedule.append(calculate_date)
                if calculate_date in self.obs_dates.date_schedule:  # 如果估值日是敲出观察日
                    obs_idx = self.obs_dates.date_schedule.index(calculate_date)
                    if observable.data >= self.barrier_out[obs_idx]:  # 如果标的价格大于等于敲出价，敲出
                        self.status = StatusType.UpTouch
                        self.stop_date = calculate_date
                        return
                if self.status in [StatusType.DownTouch, StatusType.UpTouch]:  # 如果已敲入/已敲出，维持原来的状态
                    return
                if self.status == StatusType.NoTouch:  # 如果未敲入未敲出
                    if self.in_obs_type == ExerciseType.American:  # 每日观察敲入(凤凰)
                        obs_idx = bisect.bisect_left(self.obs_dates.date_schedule, calculate_date)
                        if observable.data <= self.barrier_in[obs_idx]:  # 且标的价格小于等于敲入价，敲入
                            self.status = StatusType.DownTouch
                            return
                    elif self.in_obs_type == ExerciseType.European:  # 到期观察敲入(DCN/FCN)
                        if (calculate_date == self.end_date) and (observable.data <= self.barrier_in[-1]):
                            self.status = StatusType.DownTouch
                            return
                    else:
                        raise ValueError(f"不支持的敲入观察类型{self.in_obs_type}，仅支持美式(凤凰)/欧式(DCN/FCN)")

    def __repr__(self):
        """返回期权的描述"""
        return "FCN/DCN/Phoenix"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        t = global_evaluation_date() if t is None else t
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price

    def calc_param(self, reset_param_method: Callable, target_value: float = None, t: datetime.date = None, spot=None,
                   high=2., low=0., thres=0.001, max_iters=20, *args, **kwargs):
        """根据目标估值Y，用二分法反算衍生品参数X
        反算票息：calc_coupon = partial(calc_param, reset_param_method=self.reset_coupon)
        """
        # 若未指定估值时间和价位，则默认按期初估值反算，公允估值=名义本金(pricelib估值不含名本，因此默认target_value=期初价格s0)
        if (t is None) and (spot is None):
            t, spot = self.start_date, self.s0
            self.status = StatusType.NoTouch  # 重置期初状态
        target_value = self.s0 * self.margin_lvl if target_value is None else target_value
        return super().calc_param(reset_param_method, target_value, t, spot, high, low, thres, max_iters, *args, **kwargs)

    def reset_coupon(self, coupon, *args, **kwargs):
        """重新设置票息值，各个变种凤凰/DCN/FCN可重写此方法"""
        if self.engine.engine_type in [EngineType.PdeEngine, EngineType.QuadEngine]:
            # PDE-FDM 或 积分法
            digital_num = 0
            for sub_product in self.sub_products:
                if sub_product.__class__.__name__ == "DigitalOption":
                    digital_num += 1
            for i, sub_product in enumerate(self.sub_products[:digital_num]):
                # 二元子产品定价重新设置票息值
                sub_product.rebate = coupon * self.s0
            for j, sub_product in enumerate(self.sub_products[digital_num:-1]):
                # 欧式雪球子产品重新设置票息值
                sub_product.coupon_out = np.concatenate((np.zeros(j), np.array([coupon])))
                sub_product.coupon_div = coupon
        else:  # 蒙特卡洛模拟
            self.coupon = np.array([coupon for x in self.coupon])

    def init_sub_products(self):
        """初始化凤凰/DCN拆分成的25个子期权
        凤凰期权可以拆分为25个期权：
            拆分0：敲入70%，敲出100%，24个月不锁定，100%保证金，敲出票息为0%的经典雪球期权；敲入观察为每日观察；
                 (如果是DCN，拆分0改为欧式雪球，敲入观察为到期观察)
            拆分1：敲入70%，敲出100%，期限1个月，每月观察，0%保证金，保本雪球，1个月敲出票息为绝对0.6%，
                  未敲入未敲出票息为绝对0.6%，敲入为欧式观察；
            拆分2：敲入70%，敲出100%，期限2个月，每月观察，0%保证金，保本雪球，第1个月敲出票息为0%，第2个月敲出票息为绝对0.6%，
                  未敲入未敲出票息为绝对0.6%，敲入为欧式观察；
            ……
            拆分24：敲入70%，敲出100%，期限24个月，每月观察，0%保证金，保本雪球，前23个月敲出票息为0%，第24个月敲出票息为绝对0.6%，
                  未敲入未敲出票息为绝对0.6%，敲入为欧式观察；
        具体情形验证：
            （1）24个月内未敲入，且在某月发生敲出事件：假设第7个月敲出，拆分0敲出，收益为0；
                拆分1-6都未敲入未敲出，获得0.6%的派息；拆分7到期敲出，获得0.6%的派息；
                拆分8-24期间敲出，收益为0；客户一共获得7*0.6%派息，与凤凰期权收益一致；
            （2）24个月内未敲入（未跌破派息边界）且未敲出：拆分0收益为0，拆分1-24都未敲入未敲出，
                每个月都获得0.6%的派息，客户一共获得24*0.6%的派息，与凤凰期权收益一致；
            （3）24个月内发生敲入事件且在到期时未敲出。期权发生敲入事件，则存续期间必有部分月份跌破70%敲入线，
                假设24个月中第7和第8个月跌破70%，其余月份均在70%以上。拆分0敲入未敲出，需承担标的证券下跌的亏损；
                拆分1-6未敲入未敲出，每月获得0.6%派息；拆分7-8到期敲入，收益为0；
                拆分9-24由于是欧式观察，均未敲入未敲出，每月获得0.6%派息；
                客户需承担标的证券下跌的损失和22*0.6%的派息，与凤凰期权收益一致；
            （4）24个月内发生敲入事件且在存续期间敲出。同上，假设24个月中第7和第8个月跌破70%，并在第11个月敲出。
                拆分0敲入后敲出，收益为0；拆分1-6未敲入未敲出，每月获得0.6%派息；拆分7-8到期敲入，收益为0；
                拆分9-10未敲入未敲出，每月获得0.6%派息；拆分11到期敲出，获得0.6%派息；拆分12-24期间敲出，收益为0；
                客户一共获得9*0.6%的派息，与凤凰期权收益一致；
        在不同情形下，拆分期权的收益均与凤凰期权一致，从而拆分期权的估值之和与凤凰期权的估值相等。
        """
        # 拆分成 len(div_obs_dates)+1 个子结构
        self.sub_products = list()
        digital_num = self.lock_term - 1
        for i in range(digital_num):
            # 敲出锁定期内，相当于欧式二元期权，在派息线上方即获得月度绝对票息收益
            sub_producti = DigitalOption(start_date=self.start_date, end_date=self.div_obs_dates.date_schedule[i],
                                         strike=self.barrier_yield[i], rebate=self.coupon[i] * self.s0,
                                         exercise_type=ExerciseType.European, trade_calendar=self.trade_calendar,
                                         discrete_obs_interval=1 / self.t_step_per_year,
                                         t_step_per_year=self.t_step_per_year)
            self.sub_products.append(sub_producti)
        for j in range(len(self.obs_dates.date_schedule)):
            # 敲出观察日 & 派息观察日，相当于欧式雪球，敲入不亏损，到期未敲入获得绝对票息
            sub_coupon = np.concatenate((np.zeros(j), np.array([self.coupon[j + digital_num]])))
            sub_productj = EuropeanSnowball(s0=self.s0, start_date=self.start_date,
                                            end_date=self.div_obs_dates.date_schedule[j + digital_num],
                                            # fixme: 为了避免 div_obs_dates 与 obs_dates不一致产生的问题，暂时取派息观察日
                                            obs_dates=self.div_obs_dates.date_schedule[digital_num: j + self.lock_term],
                                            pay_dates=self.pay_dates.date_schedule[digital_num: j + self.lock_term],
                                            parti_in=0, margin_lvl=0, trade_calendar=self.trade_calendar,
                                            barrier_out=self.barrier_out[:j + 1],
                                            barrier_in=self.barrier_yield[j + digital_num],
                                            coupon_out=sub_coupon, coupon_div=self.coupon[j + digital_num],
                                            trigger=True, lock_term=digital_num + 1, status=StatusType.NoTouch,
                                            strike_lower=self.s0, engine=None, t_step_per_year=self.t_step_per_year)
            self.sub_products.append(sub_productj)
        # 每日观察敲入结构，相当于一个没有票息的保底雪球
        if self.in_obs_type == ExerciseType.American:  # 凤凰，每日观察敲入
            knock_in_snowball = FlooredSnowball
            _barrier_in = self.barrier_in[digital_num:]  # todo: 雪球目前只支持在敲出观察日改变敲入价格, 不支持锁定期内敲入价改变
        elif self.in_obs_type == ExerciseType.European:  # DCN，到期观察敲入
            knock_in_snowball = EuropeanSnowball
            _barrier_in = self.barrier_in[-1]
        else:
            raise ValueError(f"不支持的敲入观察类型{self.in_obs_type}，仅支持美式(凤凰)/欧式(DCN/FCN)")

        sub_product0 = knock_in_snowball(barrier_out=self.barrier_out, barrier_in=_barrier_in, s0=self.s0,
                                         lock_term=digital_num + 1, strike_lower=self.strike_lower,
                                         coupon_out=0, coupon_div=0, margin_lvl=self.margin_lvl, parti_in=self.parti_in,
                                         maturity=self.maturity, start_date=self.start_date, end_date=self.end_date,
                                         obs_dates=self.obs_dates.date_schedule,
                                         pay_dates=self.pay_dates.date_schedule[digital_num:],
                                         trade_calendar=CN_CALENDAR, t_step_per_year=self.t_step_per_year,
                                         engine=None, status=self.status)
        self.sub_products.append(sub_product0)
        return



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.pricing_engine_base import QuadEngine
from pricelib.common.date_handler import global_evaluation_date


class QuadPhoenixEngine(QuadEngine):
    """Phoenix/DCN/FCN 数值积分定价引擎
            较雪球式AutoCallable，增加派息价格，少了敲出和红利票息
            每个派息（敲出）观察日，如果价格高于派息价格，派固定利息，如果发生敲出，合约提前结束；
            发生敲入后，派息方式不变，到期如果未敲出，结构为看跌空头或熊市价差空头
        FCN(Fixed Coupon Note, 固定票息票据)/DCN(Dynamic Coupon Note, 动态票息票据)/Phoenix凤凰票据
            FCN：无派息价格（派息价格为0），每月固定派息；到期日观察是否敲入
            DCN：有派息价格，每个观察日如果价格高于派息价格，派固定利息；到期日观察是否敲入
            Phoenix：有派息价格，每个观察日如果价格高于派息价格，派固定利息；每日观察是否敲入
    """

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None, calc_greeks=False, s_step=0.001):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
            calc_greeks: bool，是否计算delta和gamma，默认为False，即只计算理论估值
            s_step: float，价格网格步长（百分比）
        Returns:
            若只计算pv，返回float，现值 pv
            若计算pv、delta和gamma，返回np.ndarray [down_price, pv, up_price] ，其中 down_price 和 up_price 是 s±Δs 的理论估值
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        if spot is None:
            spot = self.process.spot()

        if calc_greeks:  # 计算理论估值、delta和gamma
            value_portfolio = np.zeros(3)
        else:  # 只计算理论估值
            value_portfolio = 0

        # 截取子产品（选取end_date大于calculate_date的产品）
        sub_products = [sub_product for sub_product in prod.sub_products if sub_product.end_date >= calculate_date]
        # 子产品定价
        for i, sub_product in enumerate(sub_products):
            sub_product.engine.quad_method = self.quad_method
            sub_product.engine.n_points = self.n_points
            sub_value = sub_product.engine.calc_present_value(prod=sub_product, t=t, spot=spot,
                                                              calc_greeks=calc_greeks, s_step=s_step)
            # logging.info(f"{i}, {sub_product.__class__.__name__}, {sub_value}")
            value_portfolio += sub_value

        return value_portfolio



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from pricelib.common.utilities.enums import CallPut, InOut, UpDown, ExerciseType
from pricelib.common.pricing_engine_base import AnalyticEngine
from pricelib.common.date_handler import global_evaluation_date
from ...products.barrier.barrier_option import BarrierOption
from .analytic_barrier_engine import AnalyticBarrierEngine


class AnalyticAirbagEngine(AnalyticEngine):
    """安全气囊闭式解定价引擎
    安全气囊是美式观察向下敲入期权，敲入前是call或价差，敲入后payoff与持有标的资产涨跌幅相同(在参与率为1的情况下)
    相当于向下敲出看涨期权 + 向下触碰资产或无期权 的组合：
    1. 如果使用向下触碰资产或无期权，敲入的下跌参与率与重置后看涨参与率必须相等，因此受到限制，弃之不用
    2. 如果敲入的下跌参与率与重置后看涨参与率不相同，就需要使用执行价=期初价格的，一个向下敲入call和一个向下敲入put再组合起来
    3. 如果还存在敲入前封顶、敲入后封顶、敲入后封底，则需要再分别减去一个障碍期权，组合为带障碍的价差。
    """

    # pylint: disable=invalid-name
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        if spot is None:
            spot = self.process.spot()
        # 1. 安全气囊敲入前，payoff为向下敲出看涨期权的多头
        DOC = BarrierOption(strike=prod.strike, barrier=prod.barrier, inout=InOut.Out,
                            updown=UpDown.Down, callput=CallPut.Call, obs_type=ExerciseType.American,
                            start_date=prod.start_date, end_date=prod.end_date,
                            discrete_obs_interval=prod.discrete_obs_interval, rebate=0, floor_yield=0.0,
                            engine=AnalyticBarrierEngine(self.process))
        not_in_value = DOC.price(t=t, spot=spot)
        # 1*. 如果敲入前含有敲入前封顶价格，则敲入前的结构变为牛市价差，此时需要减去一个以敲入前封顶价格为执行价的向下敲出看涨期权
        if prod.init_strike_cap is not None:
            DOC2 = BarrierOption(strike=prod.init_strike_cap, barrier=prod.barrier, inout=InOut.Out,
                                 updown=UpDown.Down, callput=CallPut.Call, obs_type=ExerciseType.American,
                                 start_date=prod.start_date, end_date=prod.end_date,
                                 discrete_obs_interval=prod.discrete_obs_interval, rebate=0, floor_yield=0.0,
                                 engine=AnalyticBarrierEngine(self.process))
            not_in_value -= DOC2.price(t=t, spot=spot)
        # 2. 若已敲入，且标的价格上涨，这一部分payoff为向下敲入看涨期权的多头
        DIC = BarrierOption(strike=prod.s0, barrier=prod.barrier, inout=InOut.In, updown=UpDown.Down,
                            callput=CallPut.Call, obs_type=ExerciseType.American,
                            start_date=prod.start_date, end_date=prod.end_date,
                            discrete_obs_interval=prod.discrete_obs_interval, rebate=0, floor_yield=0.0,
                            engine=AnalyticBarrierEngine(self.process))
        in_upper_value = DIC.price(t=t, spot=spot)
        # 2*. 如果敲入后含有敲入后封顶价格，则敲入后的结构变为牛市价差，此时需要减去一个以敲入后封顶价格为执行价的向下敲入看涨期权
        if prod.in_strike_cap is not None:
            DIC2 = BarrierOption(strike=prod.in_strike_cap, barrier=prod.barrier, inout=InOut.In, updown=UpDown.Down,
                                 callput=CallPut.Call, obs_type=ExerciseType.American,
                                 start_date=prod.start_date, end_date=prod.end_date,
                                 discrete_obs_interval=prod.discrete_obs_interval, rebate=0, floor_yield=0.0,
                                 engine=AnalyticBarrierEngine(self.process))
            in_upper_value -= DIC2.price(t=t, spot=spot)
        # 3. 若已敲入，且标的价格下跌，这一部分payoff为向下敲入看跌期权的空头
        DIP = BarrierOption(strike=prod.s0, barrier=prod.barrier, inout=InOut.In, updown=UpDown.Down,
                            callput=CallPut.Put, obs_type=ExerciseType.American,
                            start_date=prod.start_date, end_date=prod.end_date,
                            discrete_obs_interval=prod.discrete_obs_interval, rebate=0, floor_yield=0.0,
                            engine=AnalyticBarrierEngine(self.process))
        in_lower_value = -DIP.price(t=t, spot=spot)
        # 2*. 如果敲入后含有敲入后封底价格，则敲入后的结构变为熊市价差，此时需要加上一个以敲入后封底价格为执行价的向下敲入看跌期权
        if prod.in_strike_floor is not None:
            DIP2 = BarrierOption(strike=prod.in_strike_floor, barrier=prod.barrier, inout=InOut.In, updown=UpDown.Down,
                                 callput=CallPut.Put, obs_type=ExerciseType.American,
                                 start_date=prod.start_date, end_date=prod.end_date,
                                 discrete_obs_interval=prod.discrete_obs_interval, rebate=0, floor_yield=0.0,
                                 engine=AnalyticBarrierEngine(self.process))
            in_lower_value += DIP2.price(t=t, spot=spot)
        # 4. 组合为安全气囊的价值
        value = (not_in_value * prod.call_parti  # 未敲入的看涨参与率
                 + in_upper_value * prod.reset_call_parti  # 敲入的重置后看涨参与率
                 + in_lower_value * prod.knockin_parti)  # 敲入的下跌参与率
        if prod.floor_yield != 0:  # 保底收益率
            calculate_date = global_evaluation_date() if t is None else t
            _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
            _tenor = (prod.end_date - prod.start_date).days / prod.annual_days.value  # 从起始日到终止日的自然日时间（年）
            value += prod.floor_yield * prod.s0 * _tenor * self.process.interest.disc_factor(_maturity)
        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from scipy.stats import norm

from pricelib.common.utilities.enums import CallPut, InOut, UpDown, ExerciseType, PaymentType
from pricelib.common.processes import StochProcessBase
from pricelib.common.pricing_engine_base.analytic_engine_base import (AnalyticEngine, bs_formula, euro_cash_or_nothing,
                                                                      euro_asset_or_nothing)
from pricelib.common.date_handler import global_evaluation_date
from .analytic_barrier_engine import AnalyticBarrierEngine


class AnalyticDoubleBarrierEngine(AnalyticEngine):
    """双边障碍期权闭式解定价引擎
    欧式：
        可以由欧式资产或无和欧式现金或无期权的解析解组合而成，亦支持年化的敲入/未敲出现金返还和保底收益率。
    美式：
        双边敲出期权，到期时间 T 之前，满足L<S<U，则到期回报为max(S-K,0)，否则到期回报为0。双边敲入期权等于香草期权多头与双边敲出期权空头的组合。
        注意: 下面的两种美式解析解，都不支持立即支付；都没有现金返还，即rebate=0；也不支持保底收益率，即floor_yield=0，
        Haug(1998) Put-Call Barrier Transformations 对于标的持有成本为0的期权(期货期权)，利用认购-认沽障碍期权对称性，
                                                    利用单边障碍期权解析解的组合给双障碍期权定价。
        Ikeda and Kunitomo(1992)将双障碍期权表示为加权正态分布函数的无限集，但是仅当行权价在障碍范围内时，公式才是成立的。
        Haug(1998)的解析解与Ikeda and Kunitomo(1992)的结果几乎一致，两者对比，
            Ikeda and Kunitomo(1992)适用于标的持有成本不为0的情形, Haug(1998)适用于行权价在障碍范围之外的情形。
    """

    def __init__(self, stoch_process: StochProcessBase = None, formula_type="Ikeda&Kunitomo1992",
                 series_num=10, delta1=0, delta2=0, *,
                 s=None, r=None, q=None, vol=None):
        """
        Args:
            stoch_process: 随机过程StochProcessBase对象
            self.formula_type: 解析解类型，Ikeda&Kunitomo1992或Haug1998
        其他三个参数是Ikeda&Kunitomo1992的参数：
            self.series_num: 用于计算近似解的级数项 默认 i = -10 ~ 10
            self.delta1: 上边界的曲率
            self.delta2: 下边界的曲率，
                当delta1 = delta2 = 0时，对应两条平行边界
                当delta1 < 0 < delta2 时，对应下边界随时间呈指数增长，而上百年姐随时间呈指数衰减
                当delta1 > 0 > delta2 时，对应一个向下凸的下边界和一个向上凸的上边界
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.formula_type = formula_type
        self.series_num = series_num
        self.i_vec = np.array(range(- series_num, series_num + 1)).astype(float)  # 用于计算近似解的级数项
        self.delta1 = delta1
        self.delta2 = delta2
        # 以下是计算过程的中间变量
        self.prod = None  # Product产品对象

    # pylint: disable=invalid-name, too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date,
                                                        prod.end_date) / prod.t_step_per_year  # 从估值日到到期日的交易日时间（年）
        tenor = (prod.stop_date - prod.start_date).days / prod.annual_days.value  # 从起始日到到期日的自然日时间（年）
        # 保底收益率
        _floor_yield = 0
        if prod.floor_yield != 0:  # 保底收益率需要输入期初价格s0作为名义本金
            _floor_yield = prod.floor_yield * tenor * prod.s0

        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)

        if calculate_date == prod.stop_date:  # 如果估值日就是终止日，则直接计算payoff
            value = self.calc_maturity_payoff(prod, np.array([spot]), tenor, _floor_yield)[0]
            return value

        if prod.obs_type == ExerciseType.European:  # 如果是欧式障碍期权（到期观察）
            return self._calc_european_barrier(prod, spot, tau, r, q, vol, tenor, _floor_yield)
        elif prod.obs_type == ExerciseType.American:  # 如果是美式障碍期权
            assert prod.payment_type != PaymentType.Hit, "Error: 美式双边障碍期权解析解不支持立即支付的情形，只支持到期支付。"
            assert (prod.rebate[0] == 0 and prod.rebate[1] == 0 and prod.floor_yield == 0
                    ), ("美式双边障碍期权解析解不支持现金返还和保底收益率，可以尝试使用双鲨期权解析解定价引擎，"
                        "双鲨使用了美式双边障碍期权解析解+美式双接触/双不接触解析解的组合，支持了现金返还和保底收益率")
            # 由于不支持现金返还，也就无所谓支付方式payment_type是立即支付还是到期支付了
            if prod.discrete_obs_interval is not None:
                # 均匀离散观察，M. Broadie, P. Glasserman, S.G. Kou(1997) 在连续障碍期权解析解上加调整项，调整障碍价格水平
                # 指数上的beta = -zeta(1/2) / sqrt(2*pi) = 0.5826, 其中zeta是黎曼zeta函数
                bound = [None, None]
                bound[1] = prod.bound[1] * np.exp(0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
                bound[0] = prod.bound[0] * np.exp(-0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            else:  # 连续观察
                bound = prod.bound
            if self.formula_type == "Ikeda&Kunitomo1992":
                assert (bound[0] < prod.strike < bound[1]), "Error: Ikeda and Kunitomo(1992)的解析解仅当行权价在障碍范围内时，公式才是成立的。"
                return self._Ikeda_Kunitomo1992_pv(spot, r, q, vol, tau, bound)
            if self.formula_type == "Haug1998":
                assert (r - q == 0), "Error: Haug(1998)的解析解仅当标的持有成本为0时(期货期权，或r-q=0)，公式才是准确的。"
                return self._Haug1998_pv(spot, r, q, vol, tau, bound)
            raise NotImplementedError(f"{self.formula_type}不支持的解析解类型, 仅支持Ikeda&Kunitomo1992和Haug1998")
        else:
            raise ValueError("obs_type must be European or American")

    def calc_maturity_payoff(self, prod, s_vec, _tenor, _floor_yield=0.):
        """计算终止时的期权价值，在障碍价格内侧，默认设定未敲入，未敲出
        Args:
            prod: Product产品对象
            s_vec: np.ndarray, 网格的价格格点
            _tenor: float, 从起始日到终止日的自然日时间（年）
            _floor_yield: float，处理后的保底收益绝对数值
        Returns:
            value: np.ndarray, 期末价值向量
        """
        payoff = prod.callput.value * (s_vec - prod.strike) * prod.parti
        if prod.inout == InOut.Out:
            _rebate = self._preprocess_extra_returns(prod, _tenor, prod.rebate)  # 高低敲出现金返还绝对数值
            value = np.where(s_vec <= prod.bound[0], _rebate[0],
                             np.where(s_vec >= prod.bound[1], _rebate[1],
                                      np.where(payoff < 0, 0, payoff) + _floor_yield))
        elif prod.inout == InOut.In:
            _rebate = self._preprocess_extra_returns(prod, _tenor, [prod.rebate[0], ])  # 中间的敲入现金返还绝对数值
            value = np.where((s_vec > prod.bound[0]) & (s_vec < prod.bound[1]), _rebate[0],
                             np.where(payoff < 0, 0, payoff) + _floor_yield)
        else:
            raise ValueError("InOut must be In or Out")
        return value

    @staticmethod
    def _preprocess_extra_returns(prod, period, rebate_list):
        """预处理可能存在的年化现金返还rebate
        Args:
            prod: Product产品对象
            period: float 或 np.ndarray, 从起始到期权终止的时间（年）
                    对欧式障碍、美式敲入而言是整个存续期；对美式敲出而言，期权可能会因为敲出提前终止，此时period是起始日到敲出日的时间
            rebate_list: List[float], 现金返还率或绝对数值
        Returns:
            _rebate: float，处理后的现金返还绝对数值
        """
        if prod.rebate_annual:  # 年化百分比
            _rebate = []
            for r in rebate_list:
                _rebate.append(r * period * prod.s0)
        else:
            _rebate = rebate_list  # 绝对数值
        return _rebate

    def _extra_returns(self, prod, spot, tau, r, q, vol, tenor, _floor_yield):
        """处理欧式双边障碍期权可能存在的额外payoff：现金返还rebate、保底收益率floor_yield
            若 Out敲出， 障碍两侧为现金返还rebate，障碍中间为保底收益floor_yield
            若 In敲入， 障碍两侧为保底收益floor_yield，障碍中间为现金返还rebate
        Args:
            prod: Product产品对象
            spot: float，估值日标的价格
            tau: float，交易日到期时间（年）
            r: float，无风险利率（常数）
            q: float，连续分红率（常数）
            vol: float，波动率（常数）
        Returns:
        """
        if prod.inout == InOut.Out:  # 双边敲出
            _rebate = self._preprocess_extra_returns(prod, tenor, prod.rebate)  # 高低敲出现金返还绝对数值
            extra = (euro_cash_or_nothing(s=spot, k=prod.bound[0], t=tau, r=r, q=q, sigma=vol, sign=-1) * _rebate[0]
                     + euro_cash_or_nothing(s=spot, k=prod.bound[1], t=tau, r=r, q=q, sigma=vol, sign=1) * (
                             _rebate[1] - _floor_yield)
                     + euro_cash_or_nothing(s=spot, k=prod.bound[0], t=tau, r=r, q=q, sigma=vol, sign=1)
                     * _floor_yield)
        elif prod.inout == InOut.In:  # 双边敲入
            _rebate = self._preprocess_extra_returns(prod, tenor, [prod.rebate[0], ])  # 中间的敲入现金返还绝对数值
            extra = (euro_cash_or_nothing(s=spot, k=prod.bound[0], t=tau, r=r, q=q, sigma=vol, sign=-1) * _floor_yield
                     + euro_cash_or_nothing(s=spot, k=prod.bound[1], t=tau, r=r, q=q, sigma=vol, sign=1) * (
                             _floor_yield - _rebate[0])
                     + euro_cash_or_nothing(s=spot, k=prod.bound[0], t=tau, r=r, q=q, sigma=vol, sign=1)
                     * _rebate[0])
        else:
            raise ValueError("InOut must be In or Out")
        return extra

    def _calc_european_barrier(self, prod, spot, tau, r, q, vol, tenor, _floor_yield=0.):
        """欧式双边障碍期权的解析解
            欧式香草与欧式资产或无的解析解的组合，欧式现金或无的解析解可用于复制现金返还。
        Args:
            prod: Product产品对象
            spot: float，估值日标的价格
            tau: float，交易日到期时间（年）
            r: float，无风险利率（常数）
            q: float，连续分红率（常数）
            vol: float，波动率（常数）
            tenor: float, 从起始日到终止日的自然日存续时间（年）
            _floor_yield: float, 保底收益绝对数值，计入时间为从起始日到终止日
        Returns:
            欧式双边障碍期权的解析解
        """
        # 敲出：障碍两侧获得rebate，中间获得floor_yield；敲入：障碍两侧获得floor_yield，中间获得rebate
        extra_values = self._extra_returns(prod, spot, tau, r, q, vol, tenor, _floor_yield)
        # 障碍中香草部分的收益
        if prod.inout == InOut.Out:  # 双边敲出
            if (prod.callput == CallPut.Call and prod.bound[1] <= prod.strike) or (
                    prod.callput == CallPut.Put and prod.strike <= prod.bound[0]):
                # 若香草payoff位于敲出范围，香草的收益=0，此时相当于双边二元期权 + rebate和保底收益
                value = 0 + extra_values
            elif (prod.bound[0] < prod.strike < prod.bound[1]):
                # 若执行价位于两个障碍价之间，除了rebate和保底收益之外，call/put在两个障碍中间的收益等价于如下资产组合：
                # 多一份call/put(k), 空一份方向向上/下、执行价为高障碍价/低障碍价的[资产或无-执行价*现金或无]
                if prod.callput == CallPut.Call:
                    _barrier = prod.bound[1]
                    sign = 1
                else:
                    _barrier = prod.bound[0]
                    sign = -1
                value = (bs_formula(s=spot, k=prod.strike, t=tau, r=r, q=q, sigma=vol, sign=sign) - sign *
                         (euro_asset_or_nothing(s=spot, k=_barrier, t=tau, r=r, q=q, sigma=vol, sign=sign) -
                          euro_cash_or_nothing(s=spot, k=_barrier, t=tau, r=r, q=q, sigma=vol, sign=sign) * prod.strike
                          )) * prod.parti + extra_values
            elif (prod.callput == CallPut.Call and prod.strike <= prod.bound[0]) or (
                    prod.callput == CallPut.Put and prod.bound[1] <= prod.strike):
                # 若香草payoff跨过了两个障碍价，除了rebate和保底收益之外，call/put在两个障碍外侧的收益等价于如下资产组合：
                # 多一份方向向上/下、执行价为低障碍价/高障碍价的[资产或无-执行价*现金或无]，
                # 空一份方向向上/下、执行价为高障碍价/低障碍价的[资产或无-执行价*现金或无]，
                if prod.callput == CallPut.Call:
                    _barrier1 = prod.bound[0]
                    _barrier2 = prod.bound[1]
                    sign = 1
                else:
                    _barrier1 = prod.bound[1]
                    _barrier2 = prod.bound[0]
                    sign = -1
                value = ((euro_asset_or_nothing(s=spot, k=_barrier1, t=tau, r=r, q=q, sigma=vol, sign=sign) -
                          euro_cash_or_nothing(s=spot, k=_barrier1, t=tau, r=r, q=q, sigma=vol, sign=sign) * prod.strike
                          ) -
                         (euro_asset_or_nothing(s=spot, k=_barrier2, t=tau, r=r, q=q, sigma=vol, sign=sign) -
                          euro_cash_or_nothing(s=spot, k=_barrier2, t=tau, r=r, q=q, sigma=vol, sign=sign) * prod.strike
                          )) * sign * prod.parti + extra_values
            else:
                raise ValueError(
                    f"Unhandled strike-barrier situation: strike={prod.strike}, barrier_lower={prod.bound[0]}, barrier_upper={prod.bound[1]}")
        elif prod.inout == InOut.In:  # 双边敲入
            if (prod.callput == CallPut.Call and prod.bound[1] <= prod.strike) or (
                    prod.callput == CallPut.Put and prod.strike <= prod.bound[0]):
                # 若香草payoff全部位于敲入范围，此时相当于香草的价值+rebate和保底收益
                value = bs_formula(s=spot, k=prod.strike, t=tau, r=r, q=q, sigma=vol, sign=prod.callput.value
                                   ) * prod.parti + extra_values
            elif (prod.bound[0] < prod.strike < prod.bound[1]):
                # 若执行价位于两个障碍价之间，除了rebate和保底收益之外，call/put在两个障碍外侧的收益等价于如下资产组合：
                # 多一份方向向上/下、执行价为高障碍价/低障碍价的[资产或无-执行价*现金或无]
                if prod.callput == CallPut.Call:
                    _barrier = prod.bound[1]
                    sign = 1
                else:
                    _barrier = prod.bound[0]
                    sign = -1
                value = (euro_asset_or_nothing(s=spot, k=_barrier, t=tau, r=r, q=q, sigma=vol, sign=sign) -
                         euro_cash_or_nothing(s=spot, k=_barrier, t=tau, r=r, q=q, sigma=vol, sign=sign) * prod.strike
                         ) * sign * prod.parti + extra_values
            elif (prod.callput == CallPut.Call and prod.strike <= prod.bound[0]) or (
                    prod.callput == CallPut.Put and prod.bound[1] <= prod.strike):
                # 若香草payoff跨过了两个障碍价，除了rebate和保底收益之外，call/put在两个障碍外侧的收益等价于如下资产组合：
                # 多一份call/put，
                # 空一份方向向上/下、执行价为低障碍价/高障碍价的[资产或无-执行价*现金或无]，
                # 多一份方向向上/下、执行价为高障碍价/低障碍价的[资产或无-执行价*现金或无]。
                if prod.callput == CallPut.Call:
                    _barrier1 = prod.bound[0]
                    _barrier2 = prod.bound[1]
                    sign = 1
                else:
                    _barrier1 = prod.bound[1]
                    _barrier2 = prod.bound[0]
                    sign = -1
                value = (bs_formula(s=spot, k=prod.strike, t=tau, r=r, q=q, sigma=vol, sign=sign) - sign *
                         (euro_asset_or_nothing(s=spot, k=_barrier1, t=tau, r=r, q=q, sigma=vol, sign=sign) -
                          euro_cash_or_nothing(s=spot, k=_barrier1, t=tau, r=r, q=q, sigma=vol, sign=sign) * prod.strike
                          ) + sign *
                         (euro_asset_or_nothing(s=spot, k=_barrier2, t=tau, r=r, q=q, sigma=vol, sign=sign) -
                          euro_cash_or_nothing(s=spot, k=_barrier2, t=tau, r=r, q=q, sigma=vol, sign=sign) * prod.strike
                          )) * prod.parti + extra_values
            else:
                raise ValueError(
                    f"Unhandled strike-barrier situation: strike={prod.strike}, barrier_lower={prod.bound[0]}, barrier_upper={prod.bound[1]}")
        else:
            raise ValueError("InOut must be In or Out")
        return value

    # pylint: disable=invalid-name, too-many-locals
    def _Ikeda_Kunitomo1992_pv(self, spot, r, q, vol, tau, bound):
        """Ikeda and Kunitomo(1992)级数解，仅当行权价在障碍范围内时，公式才是成立的"""
        if spot >= bound[1] or spot <= bound[0]:  # 标的价格不在障碍范围内时，直接计算payoff
            if self.prod.inout == InOut.Out:  # 发生敲出，payoff为0
                return 0
            elif self.prod.inout == InOut.In:  # 发生敲入，返回香草期权的价值
                return bs_formula(s=spot, k=self.prod.strike, t=tau, r=r, q=q, sigma=vol,
                                  sign=self.prod.callput.value) * self.prod.parti
            else:
                raise ValueError("inout只能是InOut.In或InOut.Out")

        if self.prod.callput == CallPut.Call:
            E = bound[1] * math.exp(self.delta1 * tau)
        else:
            E = bound[0] * math.exp(self.delta2 * tau)
        numerator = (r - q + 0.5 * vol ** 2) * tau
        denominator = vol * math.sqrt(tau)
        d1 = (np.log(
            spot * np.power(bound[1], 2 * self.i_vec) / (self.prod.strike * np.power(bound[0], 2 * self.i_vec)))
              + numerator) / denominator
        d2 = (np.log(spot * np.power(bound[1], 2 * self.i_vec) / (E * np.power(bound[0], 2 * self.i_vec)))
              + numerator) / denominator
        d3 = (np.log(
            np.power(bound[0], 2 * self.i_vec + 2) / (self.prod.strike * spot * np.power(bound[1], 2 * self.i_vec)))
              + numerator) / denominator
        d4 = (np.log(
            np.power(bound[0], 2 * self.i_vec + 2) / (E * spot * np.power(bound[1], 2 * self.i_vec)))
              + numerator) / denominator
        mu1 = 2 * (r - q - self.delta2 - self.i_vec * (self.delta1 - self.delta2)) / vol ** 2 + 1
        mu2 = 2 * self.i_vec * (self.delta1 - self.delta2) / vol ** 2
        mu3 = 2 * (r - q - self.delta2 + self.i_vec * (self.delta1 - self.delta2)) / vol ** 2 + 1

        Nd1a = (np.power(np.power(bound[1], self.i_vec) / np.power(bound[0], self.i_vec), mu1) * (
                bound[0] / spot) ** mu2 * (norm.cdf(d1) - norm.cdf(d2)) * self.prod.callput.value)
        Nd1b = (np.power(np.power(bound[0], self.i_vec + 1) / (np.power(bound[1], self.i_vec) * spot), mu3) * (
                norm.cdf(d3) - norm.cdf(d4)) * self.prod.callput.value)
        Nd2a = (np.power(np.power(bound[1], self.i_vec) / np.power(bound[0], self.i_vec), mu1 - 2) * (
                bound[0] / spot) ** mu2 * self.prod.callput.value * (
                        norm.cdf(d1 - denominator) - norm.cdf(d2 - denominator)))
        Nd2b = (np.power(np.power(bound[0], self.i_vec + 1) / (np.power(bound[1], self.i_vec) * spot), mu3 - 2) * (
                norm.cdf(d3 - denominator) - norm.cdf(d4 - denominator)) * self.prod.callput.value)

        double_out = (self.prod.callput.value * spot * math.exp(-q * tau) * np.sum(Nd1a - Nd1b) -
                      self.prod.callput.value * self.prod.strike * math.exp(-r * tau) * np.sum(Nd2a - Nd2b))
        if self.prod.inout == InOut.Out:
            return double_out * self.prod.parti
        elif self.prod.inout == InOut.In:
            double_in = bs_formula(s=spot, k=self.prod.strike, t=tau, r=r, q=q, sigma=vol,
                                   sign=self.prod.callput.value) - double_out
            return double_in * self.prod.parti
        else:
            raise ValueError("inout只能是InOut.In或InOut.Out")

    # pylint: disable=too-many-locals
    def _Haug1998_pv(self, spot, r, q, vol, tau, bound):
        """Haug(1998)的解析解，仅当标的持有成本为0时(期货期权，或r-q=0)，公式才是准确的"""
        value = 0
        if self.prod.callput == CallPut.Call:
            opposite_callput = CallPut.Put
        else:
            opposite_callput = CallPut.Call
        from ...products.barrier.barrier_option import BarrierOption
        barrier_analytic_engine = AnalyticBarrierEngine(self.process, for_haug=True)
        up_barrier = BarrierOption(maturity=tau, strike=self.prod.strike, barrier=bound[1],
                                   updown=UpDown.Up, callput=self.prod.callput, inout=InOut.In, rebate=0,
                                   discrete_obs_interval=None,
                                   engine=barrier_analytic_engine)
        down_barrier = BarrierOption(maturity=tau, strike=self.prod.strike, barrier=bound[0],
                                     updown=UpDown.Down, callput=self.prod.callput, inout=InOut.In, rebate=0,
                                     discrete_obs_interval=None,
                                     engine=barrier_analytic_engine)
        up_barrier_opposite = BarrierOption(strike=bound[1] ** 2 / self.prod.strike, barrier=bound[1] ** 2 / bound[0],
                                            maturity=tau, rebate=0, inout=InOut.In,
                                            updown=UpDown.Up, callput=opposite_callput,
                                            discrete_obs_interval=None,
                                            engine=barrier_analytic_engine)
        down_barrier_opposite = BarrierOption(strike=bound[0] ** 2 / self.prod.strike, barrier=bound[0] ** 2 / bound[1],
                                              maturity=tau, rebate=0, inout=InOut.In,
                                              updown=UpDown.Down, callput=opposite_callput,
                                              discrete_obs_interval=None,
                                              engine=barrier_analytic_engine)
        value += (up_barrier.price() + down_barrier.price()
                  - up_barrier_opposite.price() * self.prod.strike / bound[1]
                  - down_barrier_opposite.price() * self.prod.strike / bound[0])

        up_barrier.strike = self.prod.strike * bound[1] ** 2 / bound[0] ** 2
        up_barrier.barrier = math.pow(bound[1], 3) / bound[0] ** 2
        down_barrier.strike = self.prod.strike * bound[0] ** 2 / bound[1] ** 2
        down_barrier.barrier = math.pow(bound[0], 3) / bound[1] ** 2
        up_barrier_opposite.strike = math.pow(bound[1], 4) / (bound[0] ** 2 * self.prod.strike)
        up_barrier_opposite.barrier = math.pow(bound[1], 4) / math.pow(bound[0], 3)
        down_barrier_opposite.strike = math.pow(bound[0], 4) / (bound[1] ** 2 * self.prod.strike)
        down_barrier_opposite.barrier = math.pow(bound[0], 4) / math.pow(bound[1], 3)

        value += (up_barrier.price() * bound[0] / bound[1] + down_barrier.price() * bound[1] / bound[0]
                  - up_barrier_opposite.price() * self.prod.strike * bound[0] / bound[1] ** 2
                  - down_barrier_opposite.price() * self.prod.strike * bound[1] / bound[0] ** 2)

        up_barrier.strike = self.prod.strike * (math.pow(bound[1], 4) / math.pow(bound[0], 4))
        up_barrier.barrier = math.pow(bound[1], 5) / math.pow(bound[0], 4)
        down_barrier.strike = self.prod.strike * (math.pow(bound[0], 4) / math.pow(bound[1], 4))
        down_barrier.barrier = math.pow(bound[0], 5) / math.pow(bound[1], 4)

        value += (up_barrier.price() * bound[0] ** 2 / bound[1] ** 2
                  + down_barrier.price() * bound[1] ** 2 / bound[0] ** 2)

        double_in = min(value, bs_formula(s=spot, k=self.prod.strike, t=tau, r=r, q=q, sigma=vol,
                                          sign=self.prod.callput.value))
        if self.prod.inout == InOut.In:
            return double_in * self.prod.parti
        elif self.prod.inout == InOut.Out:
            double_out = bs_formula(s=spot, k=self.prod.strike, t=tau, r=r, q=q, sigma=vol,
                                    sign=self.prod.callput.value) - double_in
            return double_out * self.prod.parti
        else:
            raise ValueError("inout只能是InOut.In或InOut.Out")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import sys
import os
import datetime
from pricelib import *

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

set_logging_handlers(to_file=False, to_console=False)  # 关闭日志输出


def init_bsm_process(evaluation_date: datetime.date, s=100, r=0.02, q=0.0, vol=0.2):
    """设置全局估值日，并初始化一个广义BSM随机过程"""
    # 设置全局估值日
    set_evaluation_date(evaluation_date)
    # 1. 市场数据，包括标的物价格、无风险利率、分红率、波动率
    spot_price = SimpleQuote(value=s, name="测试标的价格")
    riskfree = ConstantRate(value=r, name="无风险利率")
    dividend = ConstantRate(value=q, name="测试标的贴水率")
    volatility = BlackConstVol(vol, name="测试标的波动率")
    # 2. 随机过程，BSM价格动态
    process = GeneralizedBSMProcess(spot=spot_price, interest=riskfree, div=dividend, vol=volatility)
    return process



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType
from .mc_autocallable_engine import MCAutoCallableEngine
from pricelib.common.date_handler import global_evaluation_date


class MCParisSnowballEngine(MCAutoCallableEngine):
    """巴黎雪球MC定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
                 不支持盈亏测算：输入spot = np.ndarray一维价格向量，t=终止日，输出payoff向量
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        self._maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        # 敲出观察日
        obs_dates = prod.obs_dates.count_business_days(calculate_date)
        self.obs_dates = np.array([num for num in obs_dates if num >= 0])
        # 敲入观察日
        knock_in_obs_dates = prod.knock_in_obs_dates.count_business_days(calculate_date)
        self.knock_in_obs_dates = np.round(np.array([num for num in knock_in_obs_dates if num >= 0])).astype(int)
        # 支付日 - 计算起始日到支付日的天数，用于计算应支付的敲出收益
        calculate_start_diff = (calculate_date - prod.start_date).days
        pay_dates = prod.pay_dates.count_calendar_days(prod.start_date)
        self.pay_dates = np.array([num / prod.annual_days.value for num in pay_dates if num >= calculate_start_diff])
        assert len(self.obs_dates) == len(self.pay_dates), f"Error: {prod}的观察日和付息日长度不一致"
        # 支付日 - 计算估值日到支付日的天数，用于折现
        pay_dates_tau = prod.pay_dates.count_calendar_days(calculate_date)
        self.pay_dates_tau = np.array([num / prod.annual_days.value for num in pay_dates_tau if num >= 0])
        # 经过估值日截断的列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
        self._barrier_out = prod.barrier_out[-len(self.obs_dates):].copy()
        self._coupon_out = prod.coupon_out[-len(self.obs_dates):].copy()
        self._barrier_in = prod.barrier_in[-len(self.knock_in_obs_dates):].copy()
        if spot is None:
            spot = self.process.spot()
        elif spot != self.process.spot():
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        # 当估值日就是终止日时，直接用spot计算payoff，无需随机数矩阵
        if calculate_date == prod.stop_date:
            self.s_paths = np.array([[spot, ]])  # spot是一个数值，计算期末payoff
        else:
            self.s_paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                           t_step_per_year=prod.t_step_per_year).copy()
        self._n_path = self.s_paths.shape[1]

        # 统计各个情景的payoff
        self._cal_knock_out_date()
        self._cal_knock_out_payoff()
        self._cal_knock_in_scenario(calculate_date)
        self._cal_hold_to_maturity_payoff()
        self._cal_knock_in_payoff()
        result = (self.knock_out_profit + self.hold_to_maturity_profit + self.knock_in_loss) / self._n_path
        return result


    def _cal_knock_in_scenario(self, calculate_date):
        """计算每条路径敲入时间"""
        prod = self.prod
        if prod.status == StatusType.DownTouch:
            pass
        else:
            # 判断某一条路径是否有敲入
            knock_in_level = np.tile(self._barrier_in, (self._n_path, 1)).T  # 变敲入（与敲出观察日同长度的列表）
            knock_in_bool = np.where(self.s_paths[self.knock_in_obs_dates - 1] <= knock_in_level, 1, 0)
            knock_in_count = np.cumsum(knock_in_bool, axis=0)
            knock_in_scenario = np.max(knock_in_count, axis=0) >= prod.knock_in_times
            self.knock_in_scenario = knock_in_scenario & self.not_knock_out



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress

import numpy as np

from pricelib.common.utilities.enums import StatusType, CallPut, PaymentType, OutMechanism
from pricelib.common.date_handler import (global_evaluation_date, Schedule, CN_CALENDAR, AnnualDays)
from pricelib.common.utilities.patterns import Observer, global_auto_update_status
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.mc_engines import MCAccumulatorEngine


class Accumulator(OptionBase, Observer):
    """累购累沽期权，支持标准累购累沽（敲出不终止），熔断累购累沽（敲出终止），不敲出累购累沽（无敲出）
    todo: 固定赔付累购累沽，增强型累购累沽，末日加倍累购累沽
    以累购为例：行权价通常低于二级市场标的现价，敲出障碍价通常高于二级市场标的现价。
             在每个交易日，期权买方按以下规则买入标的建仓，或者直接获得以下规则的现金收益/亏损：
                        如果标的价格在执行价之下，以行权价购买N份标的，N是杠杆比例，例如N=2，此时发生多倍建仓的亏损；
                        如果标的价格在执行价之上，以行权价购买1份标的，此时享有折价建仓的收益；
             敲出机制有三种：敲出不终止、敲出终止、不敲出。
                在敲出观察日，如果标的价格在敲出障碍价之上，发生敲出，
                        (1)敲出不终止，下一观察日继续观察。(标准累购累沽)
                        (2)敲出终止，合约提前终止 (熔断累购累沽)，一次性获得剩余观察日（含敲出当日）的每日固定现金补偿。
                (3)不敲出，只要标的价格在行权价之上，就可以用行权价购买1份标的。但是不敲出的累购的行权价一般高于标准累沽，折价建仓收益较小。
             到期时，最终的收益即是累计的所有标的份数 x (结算时的标的价格 - 行权价)。
    累沽与累沽相反，行权价通常高于二级市场标的现价，敲出障碍价通常低于二级市场标的现价。
    """

    def __init__(self, s0, amt, barrier_out=None, strike=None, leverage_ratio=None, callput=CallPut.Call,
                 out_mechanism=OutMechanism.NoTerminate, rebate=0.0, payment_type=PaymentType.Expire, obs_type="d",
                 engine=None, status=(0.0, StatusType.NoTouch), maturity=None, start_date=None, obs_dates=None,
                 end_date=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 stoch_process=None, s=None, r=None, q=None, vol=None, engine_method=None):
        """构造函数
        产品参数:
            s0: float，标的初始价格
            amt: float，单日执行量，例如10吨、100手等
            callput: CallPut枚举类，期权类型，call累购/put累沽
            strike: float，行权价
            leverage_ratio: float, 杠杆比例，累购期权下，当标的价格在执行价之下时，购买leverage_ratio份标的；
                                                    当标的价格在执行价之上、敲出价之下时，购买1份标的。累沽同理。
            out_mechanism: OutMechanism枚举类，敲出机制，有三种：敲出不终止NoTerminate/敲出终止（熔断）Terminate/不敲出NoKnockOut
            barrier_out: float，敲出障碍价
            rebate: float，敲出补偿(元/(天*amt))。发生敲出时，合约提前终止（敲出熔断），一次性获得剩余观察日（含敲出当日）的每日固定现金补偿。
            payment_type: PaymentType枚举类，结算方式，有两种：每个观察日建仓，到期结算Expire/每个观察日立即结算现金收益Hit
            obs_type: str，观察日类型，仅支持每日观察 'd'
            status: 存续期的持仓状态。(amount, isknockout), isknockout代表是否发生敲出，默认为StatusType.NoTouch未敲出
                    当PaymentType.Expire时，amount代表已累计仓位，正数表示买入，负数表示卖出，期初默认是0.
                    当PaymentType.Hit时，amount代表已发生的收益/亏损的现值。期初默认是0.
        时间参数: 要么输入年化期限，要么输入起始日和到期日；敲出观察日可缺省
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_dates: List[datetime.date]，敲出观察日，可缺省，缺省时会自动生成每月观察的敲出日期序列(已根据节假日调整)
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            engine: 定价引擎，PricingEngine类，仅支持Monte Carlo模拟 MCAccumulatorEngine
            engine_method: str, 定价引擎种类。若未输入engine，可以用engine_method快捷生成指定定价引擎。可选MC/MonteCarlo
            若未提供引擎engine、也未指定引擎种类engine_method的情况下，如果提供了 随机过程 或 标的价格、无风险利率、分红/融券率、波动率，
            则默认使用蒙特卡洛模拟定价引擎 (BSM模型)
            stoch_process: 随机过程StochProcessBase对象
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.s0 = s0  # 标的初始价格
        self.amt = amt  # 单日执行量
        self.trade_calendar = trade_calendar  # 交易日历
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if self.trade_calendar.isholiday(self.end_date):  # 如果到期日不是交易日，向后顺延一个交易日
            self.end_date = self.trade_calendar.advance(self.end_date, datetime.timedelta(days=1))
        if maturity is None:
            self.maturity = (end_date - start_date).days / annual_days.value
        else:
            self.maturity = maturity
        self.callput = callput  # 期权类型: 累购/累沽
        self.out_mechanism = out_mechanism  # 敲出机制
        self.barrier_out = barrier_out  # 敲出线
        self.strike = strike  # 行权价
        self.payment_type = payment_type  # 结算方式
        self.leverage_ratio = leverage_ratio  # 杠杆比率
        if self.out_mechanism != OutMechanism.NoKnockOut:
            if self.callput == CallPut.Call:  # 累购
                assert barrier_out > strike, "Error:有敲出的累购期权的敲出障碍价必须大于行权价。"
            else:  # 累沽
                assert barrier_out < strike, "Error:有敲出的累沽期权的敲出障碍价必须小于行权价。"
        if obs_dates is None:  # 观察日列表
            self.obs_dates = Schedule(trade_calendar=trade_calendar, start=self.start_date, end=self.end_date,
                                      freq=obs_type, lock_term=0)  # 累购累沽无锁定期
            self.end_date = self.obs_dates.end  # 修正结束时间
        else:
            self.obs_dates = Schedule(trade_calendar=trade_calendar, date_schedule=obs_dates)
        self.status = list(status)
        self.rebate = rebate
        if engine is not None:
            self.set_pricing_engine(engine)
        else:
            if engine_method is None:
                default_engine = MCAccumulatorEngine  # 默认MC定价引擎
            elif engine_method in ["MC", "MonteCarlo"]:
                default_engine = MCAccumulatorEngine
            else:
                raise NotImplementedError(f"不支持的engine_method{engine_method}, {self.__repr__()}仅支持MC")
            if stoch_process is not None:
                self.set_pricing_engine(default_engine(stoch_process))
            elif s is not None and r is not None and q is not None and vol is not None:
                self.set_pricing_engine(default_engine(s=s, r=r, q=q, vol=vol))

    def set_pricing_engine(self, engine):
        """切换定价引擎"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新状态status"""
        if global_auto_update_status():
            calculate_date = global_evaluation_date()
            if calculate_date < self.start_date or calculate_date > self.end_date:
                return  # 如果在起始日前或到期日之后，状态维持不变
            if observable == self.engine.process.spot:
                if calculate_date in self.obs_dates.date_schedule:  # 如果估值日是敲出观察日
                    self._accumulating(observable.data, calculate_date)

    def _accumulating(self, new_price, calculate_date):
        """根据最新价格，判断收益支付"""
        if (self.callput == CallPut.Call and new_price < self.strike) or (  # 累购
                self.callput == CallPut.Put and new_price > self.strike):  # 累沽
            self.discounted_positioning(new_price, self.leverage_ratio)  # 折价建仓
        else:  # 标的价格没有以不利方向突破执行价时
            if self.out_mechanism == OutMechanism.NoKnockOut:  # 不敲出
                self.discounted_positioning(new_price, 1)  # 折价建仓
            else:  # 有敲出价格
                obs_idx = self.obs_dates.date_schedule.index(calculate_date)
                if isinstance(self.barrier_out, (float, int)):
                    barrier_out = self.barrier_out
                elif isinstance(self.barrier_out, (list, np.ndarray)):
                    barrier_out = self.barrier_out[obs_idx]
                else:
                    raise TypeError(f"Error: 不支持的敲出障碍价类型{type(self.barrier_out)}")
                if self.out_mechanism == OutMechanism.Terminate:  # 敲出终止（熔断）
                    if (self.callput == CallPut.Call and new_price > barrier_out):  # 累购
                        self.status[1] = StatusType.UpTouch
                        logging.info(f"{self}已向上敲出，提前终止")
                    elif (self.callput == CallPut.Put and new_price < barrier_out):  # 累沽
                        self.status[1] = StatusType.DownTouch
                        logging.info(f"{self}已向下敲出，提前终止")
                    else:
                        self.discounted_positioning(new_price, 1)  # 折价建仓
                elif self.out_mechanism == OutMechanism.NoTerminate:  # 敲出不终止
                    if (self.callput == CallPut.Call and new_price > barrier_out) or (  # 累购
                            self.callput == CallPut.Put and new_price < barrier_out):  # 累沽
                        pass
                    else:
                        self.discounted_positioning(new_price, 1)  # 折价建仓
                else:
                    raise NotImplementedError(f"不支持的敲出机制{self.out_mechanism}")

    def discounted_positioning(self, new_price, ratio=1):
        """观察日，折价建仓/立即结算现金收益"""
        if self.payment_type == PaymentType.Expire:  # 折价建仓
            new_amt = self.callput.value * self.amt * ratio
            self.status[0] += new_amt
            logging.info(f"{self}新增建仓量{new_amt}，累计建仓量{self.status[0]}")
        else:  # 立即结算现金收益
            new_pnl = (new_price - self.strike) * self.callput.value * self.amt * ratio
            self.status[0] = round(self.status[0] + new_pnl, 4)
            logging.info(f"{self} 新增现金盈亏 {new_pnl}，累计现金盈亏 {self.status[0]}")

    def __repr__(self):
        """返回期权的描述"""
        if self.callput == CallPut.Call:
            callput_str = "累购"  # (Accumulator)
        else:
            callput_str = "累沽"  # (Decumulator)
        if self.out_mechanism == OutMechanism.NoTerminate:
            out_mechanism_str = "标准"
        elif self.out_mechanism == OutMechanism.Terminate:
            out_mechanism_str = "熔断"
        else:
            out_mechanism_str = "不敲出"
        if self.payment_type == PaymentType.Expire:
            payment_type_str = "建仓到期结算"
        else:
            payment_type_str = "现金立即结算"
        return f"{out_mechanism_str}{callput_str}-{payment_type_str}"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from functools import lru_cache
import numpy as np
from pricelib.common.utilities.enums import UpDown, InOut, PaymentType, ExerciseType
from pricelib.common.utilities.patterns import HashableArray
from pricelib.common.date_handler import global_evaluation_date
from pricelib.common.pricing_engine_base.quad_engine_base import QuadEngine
from pricelib.common.pricing_engine_base.analytic_engine_base import euro_cash_or_nothing, euro_asset_or_nothing


class QuadBarrierEngine(QuadEngine):
    """障碍期权数值积分定价引擎
    只支持离散观察(默认为每日观察)；敲入现金返还为到期支付；敲出现金返还支持 立即支付/到期支付"""

    @lru_cache(maxsize=1)
    def set_quad_price_range(self, s_sliced, barrier, updown):
        """
        设置数值积分的区间(价格范围)
        Args:
            s_sliced: np.ndarray = self.s[:, step], 其中step是从后向前的步数
            barrier: float, 障碍价格
            updown: 向上/向下，UpDown枚举类
        Returns:
            quad_range: tuple, (np.ndarray, ), 数值积分的区间(索引)
            an_range: tuple, (np.ndarray, ), BSM解析式资产或无N(d1)、现金或无N(d2)计算积分的区间(索引)
        """
        if updown == UpDown.Up:
            quad_range = np.where(s_sliced < barrier)
            an_range = np.where(s_sliced >= barrier)
        elif updown == UpDown.Down:
            quad_range = np.where(s_sliced > barrier)
            an_range = np.where(s_sliced <= barrier)
        else:
            raise ValueError("updown must be Up or Down")
        return quad_range, an_range

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None, calc_greeks=False, s_step=0.001):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
            calc_greeks: bool，是否计算delta和gamma，默认为False，即只计算理论估值
            s_step: float，价格网格步长（百分比）
        Returns:
            若只计算pv，返回float，现值 pv
            若计算pv、delta和gamma，返回np.ndarray [down_price, pv, up_price] ，其中 down_price 和 up_price 是 s±Δs 的理论估值
        """
        if prod.obs_type == ExerciseType.European:
            raise NotImplementedError("障碍期权积分法定价引擎暂不支持欧式障碍期权（到期观察），只支持美式障碍期权")
        if prod.rebate_annual:
            raise NotImplementedError("障碍期权积分法定价引擎暂不支持年化的现金返还收益率，请输入固定的现金返还的绝对数值")  # todo
        if prod.floor_yield != 0:
            raise NotImplementedError("障碍期权积分法定价引擎暂不支持保底收益率")  # todo
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        if tau == 0:  # 如果估值日就是到期日
            if calc_greeks:  # 计算理论估值、delta和gamma
                value = self._init_terminal_condition(prod, np.array([1 - s_step, 1, 1 + s_step]) * spot)
            else:  # 只计算理论估值, 直接计算终止条件
                value = self._init_terminal_condition(prod, np.array([spot]))[0]
            return value

        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)
        dt = 1 / prod.t_step_per_year if prod.discrete_obs_interval is None else prod.discrete_obs_interval
        self.backward_steps = int(tau / dt)
        self._check_method_params()

        # v_grid = np.empty(shape=(self.n_points, self.backward_steps + 1))
        # 设置积分法engine参数
        self.set_quad_params(r=r, q=q, vol=vol)
        # 初始化fft的对数价格向量及边界的对数价格向量
        self.init_grid(spot, vol, tau)
        s_vec = np.exp(self.ln_s_vec)

        # 从后向前进行积分计算
        # 设定期末价值，在障碍价格内侧，默认设定未敲入，未敲出
        v_vec = self._init_terminal_condition(prod, s_vec)
        for step in range(self.backward_steps - 1, 0, -1):
            quad_range, an_range = self.set_quad_price_range(s_vec, prod.barrier, prod.updown)
            # 数值积分计算区域
            y = self.ln_s_vec
            x = self.ln_s_vec[quad_range]
            v_vec[quad_range] = self.fft_step_backward(x, y, v_vec, dt)
            # v_grid[:, step] = np.minimum(v_grid[:, step], 1e250)  # 防止价格极大和极小时，期权价值趋近于无穷大
            # 解析公式计算积分区域(论文中的Vma、Vmb，相当于资产或无、现金或无，BSM的两部分)
            if prod.inout == InOut.In:
                x = HashableArray(self.ln_s_vec[an_range])
                v_vec[an_range] = euro_asset_or_nothing(s=np.exp(x), k=prod.strike, t=dt * (self.backward_steps - step),
                                                        r=self.r, q=self.q, sigma=self.vol, sign=prod.callput.value
                                                        ) * prod.callput.value * prod.parti
                v_vec[an_range] -= euro_cash_or_nothing(s=np.exp(x), k=prod.strike, t=dt * (self.backward_steps - step),
                                                        r=self.r, q=self.q, sigma=self.vol, sign=prod.callput.value
                                                        ) * prod.strike * prod.callput.value * prod.parti
            elif prod.inout == InOut.Out:
                if prod.payment_type == PaymentType.Expire:  # 到期支付现金补偿
                    v_vec[an_range] = prod.rebate * np.exp(-r * dt * (self.backward_steps - step))
                else:  # 立即支付现金补偿
                    v_vec[an_range] = prod.rebate
            else:
                raise ValueError("inout must be In or Out")

        if calc_greeks:  # 计算理论估值、delta和gamma
            up_price = self.fft_step_backward(np.array([np.log(spot * (1 + s_step))]), self.ln_s_vec, v_vec, dt)[0]
            pv = self.fft_step_backward(np.array([np.log(spot)]), self.ln_s_vec, v_vec, dt)[0]
            down_price = self.fft_step_backward(np.array([np.log(spot * (1 - s_step))]), self.ln_s_vec, v_vec, dt)[0]
            return np.array([down_price, pv, up_price])
        else:  # 只计算理论估值
            x = np.log(np.array([spot]))
            value = self.fft_step_backward(x, self.ln_s_vec, v_vec, dt)[0]
            return value

    @staticmethod
    def _init_terminal_condition(prod, s_vec):
        """初始化终止时的期权价值，在障碍价格内侧，默认设定未敲入，未敲出
        Args:
            prod: Product产品对象
            s_vec: np.ndarray, 网格的价格格点
        Returns:
            payoff: np.ndarray, 期末价值向量
        """
        if prod.inout == InOut.Out:
            v_vec = np.maximum(prod.callput.value * (s_vec - prod.strike), 0) * prod.parti
            if prod.updown == UpDown.Up:
                v_vec = np.where(s_vec > prod.barrier, prod.rebate, v_vec)
            elif prod.updown == UpDown.Down:
                v_vec = np.where(s_vec < prod.barrier, prod.rebate, v_vec)
            else:
                raise ValueError("updown must be Up or Down")
        elif prod.inout == InOut.In:
            v_vec = prod.rebate * np.ones(len(s_vec))
            if prod.updown == UpDown.Up:
                v_vec = np.where(s_vec > prod.barrier, np.maximum(prod.callput.value * (s_vec - prod.strike),
                                                                  0) * prod.parti, v_vec)
            elif prod.updown == UpDown.Down:
                v_vec = np.where(s_vec < prod.barrier, np.maximum(prod.callput.value * (s_vec - prod.strike),
                                                                  0) * prod.parti, v_vec)
            else:
                raise ValueError("updown must be Up or Down")
        else:
            raise ValueError("inout must be In or Out")
        return v_vec



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.date_handler import global_evaluation_date
from pricelib.common.pricing_engine_base import McEngine


class MCRangeAccuralEngine(McEngine):
    """区间累计 Monte Carlo 模拟定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
                 如果输入spot = np.ndarray一维价格向量，t=终止日，则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)

        if spot is None:
            spot = self.process.spot()
        # 当估值日就是到期日时，直接用spot计算payoff，无需随机数矩阵
        if calculate_date == prod.end_date:
            if isinstance(spot, (list, np.ndarray)):  # spot是一个np.ndarray一维价格向量，返回盈亏测算向量，即每个价格对应的现值，而不是平均值
                paths = np.array([spot, ])
            else:  # spot是一个数值，计算期末payoff
                paths = np.array([[spot, ]])
        else:
            if spot != self.process.spot():
                self.reset_paths_flag()  # 重置路径标志位，重新生成路径
            paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                        t_step_per_year=prod.t_step_per_year).copy()
        _n_path = paths.shape[1]

        # 敲出与行权日期(paths[1:]计算时排除已确定的估值日价格，估值之前，该日的状态应该已经随着价格的变化，更新到了status中)
        n_coupon = np.sum((paths[1:] < prod.strike_upper) & (paths[1:] > prod.strike_lower), axis=0) + prod.status
        # 支付 = 名义本金 * 年化收益 * 计息天数/365 * 位于区间内的观察日天数/观察日总天数
        payoff = (prod.s0 * prod.payment * _maturity * (n_coupon / _maturity_business_days)
                  * self.process.interest.disc_factor(_maturity))

        if (calculate_date == prod.end_date) and isinstance(spot, np.ndarray):
            return payoff  # 终止日, spot是价格向量, 则返回盈亏测算向量，即每个价格对应的现值，而不是平均值
        else:
            return np.mean(payoff)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
from numba import njit
from ..utilities.enums import ProcessType
from ..utilities.patterns import Observable, Observer
from .stoch_process import StochProcessBase


@njit(parallel=True, fastmath=True, cache=True)
def evolve_bs(t0, x0, dt, dw, drift, diffusion):
    """Black-Scholes-Merton SDE演化函数，jit加速
    Black-Scholes-Merton SDE: dS/S = drift * dt + diffusion * dW
    Args:
        t0: float, 距离起始日的年化时间
        x0: np.ndarray, t0时刻的标的价格向量
        dt: float, 年化时间增量
        dw: np.ndarray, 布朗运动增量，这里只是标准正态分布随机数，函数内部会 * sqrt(dt)
        drift: float, BSM的漂移率
        diffusion: np.ndarray, BSM的波动率
    Returns:
        np.ndarray, t0+dt时刻的标的价格向量
    """
    return x0 * (1 + drift * dt + diffusion * dw * math.sqrt(dt))


class GeneralizedBSMProcess(StochProcessBase, Observer, Observable):
    """BS风险中性 SDE: dS/S = (r-q)dt + vol dW
    既是观察者又是被观察者，观察者模式的中介者。观察S、r、q、sigma的变化，被定价引擎engine观察。
    """
    process_type = ProcessType.BSProcess1D

    def __init__(self, spot: Observable, interest: Observable, div: Observable, vol: Observable):
        """标的价格、无风险利率、分红融券率、波动率，都是被观察者。
        vol可以是常数波动率，也可以是局部波动率，可以由vol(t,S)取值"""
        self.spot = spot
        self.interest = interest
        self.div = div
        self.vol = vol
        spot.add_observer(self)  # 将本实例加入spot观察者列表
        interest.add_observer(self)  # 将本实例加入interest观察者列表
        div.add_observer(self)  # 将本实例加入div观察者列表
        vol.add_observer(self)  # 将本实例加入vol观察者列表
        super().__init__()  # 调用被观察者Observable的构造函数

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.spot.remove_observer(self)
        self.interest.remove_observer(self)
        self.div.remove_observer(self)
        self.vol.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """被观察者的值发生变化时，自动调用update方法，通知观察者"""
        self.notify_observers()  # 自动向观察者发送通知, 参数已改变

    def __call__(self):
        return ProcessType.BSProcess1D

    def drift(self, t):
        """BSM过程的漂移项
        Args:
            t: float，距离起始日的年化时间
        Returns:
            drift: float = r - q，r是无风险利率，q是分红融券率
        """
        return self.interest(t) - self.div(t)

    def diffusion(self, t, spot):
        """BSM过程的扩散项
        Args:
            t: float，距离起始日的年化时间
            spot: np.ndarray，标的价格向量
        Returns:
            diffusion: np.ndarray = vol，vol是波动率
        """
        return self.vol(t, spot)

    def evolve(self, t, x, dt, dw):
        """演化函数，根据BSM SDE进行演化
        Args:
            t: float，距离起始日的年化时间
            x: np.ndarray, t0时刻的标的价格向量
            dt: float, 年化时间增量
            dw: np.ndarray, 布朗运动增量，这里只是标准正态分布随机数，函数内部会 * sqrt(dt)
        Returns:
            np.ndarray, t0+dt时刻的标的价格向量
        """
        return evolve_bs(t, x, dt, dw, self.drift(t), self.diffusion(t, x))

    def get_fn_pde_coef(self, maturity, spot):
        """获取返回pde系数的函数
        Args:
            maturity: float，年化到期时间
            spot: float，标的价格
        Returns:
            fn_pde_coef: function, 返回PDE系数组件abc的函数
        """

        # TODO: 目前局部波动率使用K坐标，以后可能重构为log moneyness坐标
        # def fn_pde_coef(u, z):  # z在校准中是执行价K，在定价中是标的价S_u, u是距离起始日的年化时间
        #     """返回PDE系数组件abc的函数，设pde有限差分的价格向量索引为i_vec，
        #     a是需要乘以i_vec^2的系数，b是需要乘以i_vec的系数，c是不需要乘以i_vec的系数"""
        #     r_u = self.interest(u)  # u时刻的无风险利率r
        #     q_u = self.div(u)  # u时刻的分红融券率q
        #     fwd_u = spot * math.exp((maturity - u) * (r_u - q_u))  # u时刻的远期价格
        #     lv = self.vol(u, np.log(z / fwd_u))  # u时刻，log(K/F)对数moneyness的local_vol
        #     a, b, c = lv ** 2, r_u - q_u, r_u
        #     return a, b, c
        def fn_pde_coef(t, spot):  # z在校准中是执行价K，在定价中是标的价S_u, u是距离起始日的年化时间
            """返回PDE系数组件abc的函数
            Args:
                t: float，年化到期时间
                spot: float，标的价格
            Returns: 设pde有限差分的价格向量索引为i_vec
                a: 需要乘以i_vec^2的系数
                b: 需要乘以i_vec的系数
                c: 不需要乘以i_vec的系数
            """
            r_t = self.interest(t)  # t时刻的无风险利率r
            q_t = self.div(t)  # t时刻的分红融券率q
            lv = self.vol(t, spot)  # t时刻，spot的local_vol
            a, b, c = lv ** 2, r_t - q_t, r_t
            return a, b, c

        return fn_pde_coef
